"""Electrophys tool."""

import logging
import tempfile
from statistics import mean
from typing import Any, ClassVar, Literal

import entitysdk.client
from bluepyefe.extract import extract_efeatures
from efel.units import get_unit
from entitysdk.models import ElectricalCellRecording
from fastapi import HTTPException
from pydantic import BaseModel, Field

from obi_one.core.block import Block
from obi_one.core.form import Form
from obi_one.core.single import SingleCoordinateMixin

POSSIBLE_PROTOCOLS = {
    "idrest": ["idrest"],
    "idthresh": ["idthres", "idthresh"],
    "iv": ["iv"],
    "apwaveform": ["apwaveform"],
    "spontaneous": ["spontaneous"],
    "step": ["step"],
    "spontaps": ["spontaps"],
    "firepattern": ["firepattern"],
    "sponnohold30": ["sponnohold30", "spontnohold30"],
    "sponthold30": ["sponhold30", "sponthold30"],
    "starthold": ["starthold"],
    "startnohold": ["startnohold"],
    "delta": ["delta"],
    "sahp": ["sahp"],
    "idhyperpol": ["idhyeperpol"],
    "irdepol": ["irdepol"],
    "irhyperpol": ["irhyperpol"],
    "iddepol": ["iddepol"],
    "ramp": ["ramp"],
    "apthresh": ["apthresh", "ap_thresh"],
    "hyperdepol": ["hyperdepol"],
    "negcheops": ["negcheops"],
    "poscheops": ["poscheops"],
    "spikerec": ["spikerec"],
    "sinespec": ["sinespec"],
}
POSSIBLE_PROTOCOLS_STR = "', '".join(POSSIBLE_PROTOCOLS.keys())

EFEL_SETTINGS = {"strict_stiminterval": True, "Threshold": -20.0, "interp_step": 0.025}

STIMULI_TYPES = list[
    Literal[
        "spontaneous",
        "idrest",
        "idthres",
        "apwaveform",
        "iv",
        "step",
        "spontaps",
        "firepattern",
        "sponnohold30",
        "sponhold30",
        "starthold",
        "startnohold",
        "delta",
        "sahp",
        "idhyperpol",
        "irdepol",
        "irhyperpol",
        "iddepol",
        "ramp",
        "ap_thresh",
        "hyperdepol",
        "negcheops",
        "poscheops",
        "spikerec",
        "sinespec",
    ]
]

STEP_LIKE_STIMULI_TYPES = list[
    Literal[
        "idrest",
        "idthresh",
        "apwaveform",
        "iv",
        "step",
        "firepattern",
        "delta",
    ]
]

CALCULATED_FEATURES = list[
    Literal[
        "spike_count",
        "time_to_first_spike",
        "time_to_last_spike",
        "inv_time_to_first_spike",
        "doublet_ISI",
        "inv_first_ISI",
        "ISI_log_slope",
        "ISI_CV",
        "irregularity_index",
        "adaptation_index",
        "mean_frequency",
        "strict_burst_number",
        "strict_burst_mean_freq",
        "spikes_per_burst",
        "AP_height",
        "AP_amplitude",
        "AP1_amp",
        "APlast_amp",
        "AP_duration_half_width",
        "AHP_depth",
        "AHP_time_from_peak",
        "AP_peak_upstroke",
        "AP_peak_downstroke",
        "voltage_base",
        "voltage_after_stim",
        "ohmic_input_resistance_vb_ssse",
        "steady_state_voltage_stimend",
        "sag_amplitude",
        "decay_time_constant_after_stim",
        "depol_block_bool",
    ]
]


class AmplitudeInput(BaseModel):
    """Amplitude class."""

    min_value: float | None = Field(default=None, description="Minimum amplitude (nA)")
    max_value: float | None = Field(default=None, description="Maximum amplitude (nA)")


class ElectrophysiologyMetricsForm(Form):
    """Form for extracting electrophysiological metrics from a trace."""

    single_coord_class_name: ClassVar[str] = "ElectrophysiologyMetrics"
    name: ClassVar[str] = "Electrophysiology Metrics"
    description: ClassVar[str] = "Calculates ephys metrics for a given trace."

    class Initialize(Block):
        trace_id: str = Field(
            description="ID of the trace of interest."
        )
        protocols: STIMULI_TYPES | None = Field(
            default=None,
            description=f"Type of stimuli requested by the user. Should be one of: '{POSSIBLE_PROTOCOLS_STR}'."
        )
        requested_metrics: CALCULATED_FEATURES | None = Field(
            default=None,
            description="Feature requested by the user. Should be one of 'spike_count',"
            "'time_to_first_spike', 'time_to_last_spike',"
            "'inv_time_to_first_spike', 'doublet_ISI', 'inv_first_ISI',"
            "'ISI_log_slope', 'ISI_CV', 'irregularity_index', 'adaptation_index',"
            "'mean_frequency', 'strict_burst_number', 'strict_burst_mean_freq',"
            "'spikes_per_burst', 'AP_height', 'AP_amplitude', 'AP1_amp', 'APlast_amp',"
            "'AP_duration_half_width', 'AHP_depth', 'AHP_time_from_peak',"
            "'AP_peak_upstroke', 'AP_peak_downstroke', 'voltage_base',"
            "'voltage_after_stim', 'ohmic_input_resistance_vb_ssse',"
            "'steady_state_voltage_stimend', 'sag_amplitude',"
            "'decay_time_constant_after_stim', 'depol_block_bool'"
        )
        amplitude: AmplitudeInput | None = Field(
            default=None,
            description=(
            "Amplitude of the protocol (should be specified in nA)."
            "Can be a range of amplitudes with min and max values"
            "Can be None (if the user does not specify it)"
            " and all the amplitudes are going to be taken into account."
        ),
    )

    initialize: Initialize


class ElectrophysiologyMetricsOutput(BaseModel):
    """Output schema for electrophysiological metrics extracted using BluePyEfe."""

    feature_dict: dict[str, dict[str, Any]] = Field(
        description="Mapping of feature name to its metric values. "
                    "Each entry contains at least an 'avg', and optionally 'unit', 'num_traces', etc."
    )

    @classmethod
    def from_efeatures(cls, raw: dict[str, Any]) -> "ElectrophysiologyMetricsOutput":
        return cls(feature_dict=raw)


class ElectrophysiologyMetrics(ElectrophysiologyMetricsForm, SingleCoordinateMixin):
    def run(self, db_client: entitysdk.client.Client = None):
        try:
            ephys_metrics = get_electrophysiology_metrics(
                trace_id=self.initialize.trace_id,
                entity_client=db_client,
                calculated_feature=self.initialize.requested_metrics,
                amplitude=self.initialize.amplitude,
                stimuli_types=self.initialize.protocols,
            )
            return ephys_metrics
        except Exception as e:
            raise HTTPException(status_code=500, detail=f"Internal Server Error: {e}")

def get_electrophysiology_metrics(  # noqa: PLR0914, C901
    trace_id: str,
    entity_client: entitysdk.client.Client,
    calculated_feature: CALCULATED_FEATURES | None = None,
    amplitude: AmplitudeInput | None = None,
    stimuli_types: STIMULI_TYPES | None = None,
) -> ElectrophysiologyMetricsOutput:
    """Compute electrophys features for a given trace."""
    logger = logging.getLogger(__name__)

    logger.info(
        "Entering electrophys tool. Inputs: trace_id=%r, calculated_feature=%r, amplitude=%r, stimuli_types=%r",
        trace_id,
        calculated_feature,
        amplitude,
        stimuli_types,
    )

    # Deal with cases where user did not specify stimulus type or/and feature
    if not stimuli_types:
        # Default to all protocol types if not specified
        logger.warning("No stimulus type specified. Iterating over all STEP_LIKE_STIMULI_TYPES.")
        stimuli_types = list(STEP_LIKE_STIMULI_TYPES.__args__[0].__args__)  # type: ignore

    if not calculated_feature:
        # Compute ALL of the available features if not specified
        logger.warning("No feature specified. Defaulting to everything.")
        calculated_feature = list(CALCULATED_FEATURES.__args__[0].__args__)  # type: ignore

    # Turn amplitude requirement of user into a bluepyefe compatible representation
    if (
        isinstance(amplitude, AmplitudeInput) and
        amplitude.min_value is not None and
        amplitude.max_value is not None
    ):
        # If the user specified amplitude/a range of amplitudes,
        # the target amplitude is centered on the range and the
        # tolerance is set as half the range
        desired_amplitude = mean(
            [
                amplitude.min_value,
                amplitude.max_value,
            ]
        )

        # If the range is just one number, use 10% of it as tolerance
        if amplitude.min_value == amplitude.max_value:
            desired_tolerance = amplitude.max_value * 0.1
        else:
            desired_tolerance = amplitude.max_value - desired_amplitude
    else:
        # If the amplitudes are not specified, take an arbitrarily high tolerance
        desired_amplitude = 0
        desired_tolerance = 1e12
    logger.info(
        "target amplitude set to %s nA. Tolerance is %s nA",
        desired_amplitude,
        desired_tolerance,
    )

    targets = []

    for stim_type in stimuli_types:
        for efeature in calculated_feature:
            for protocol in POSSIBLE_PROTOCOLS[stim_type]:
                target = {
                    "efeature": efeature,
                    "protocol": protocol,
                    "amplitude": desired_amplitude,
                    "tolerance": desired_tolerance,
                }
                targets.append(target)
    logger.info("Generated %d targets.", len(targets))
    logger.info("Trace ID: %s", trace_id)
    trace_metadata = entity_client.get_entity(
        entity_id=trace_id, entity_type=ElectricalCellRecording
    )
    # Download the .nwb file associated to the trace from the KG
    with (
        tempfile.NamedTemporaryFile(suffix=".nwb") as temp_file,
        tempfile.TemporaryDirectory() as temp_dir,
    ):
        for asset in trace_metadata.assets:
            logger.debug("Asset object: %s", asset)
            if asset.content_type == "application/nwb":
                trace_content = entity_client.download_content(
                    entity_id=trace_id, entity_type=ElectricalCellRecording, asset_id=asset.id
                )
                temp_file.write(trace_content)
                temp_file.flush()
                break
        else:
            raise ValueError(
                f"No asset with content type 'application/nwb' found for trace {trace_id}."
            )

        # LNMC traces need to be adjusted by an output voltage of 14mV due to their experimental protocol
        files_metadata = {
            "test": {
                stim_type: [
                    {
                        "filepath": temp_file.name,
                        "protocol": protocol,
                        "ljp": trace_metadata.ljp,
                    }
                    for protocol in POSSIBLE_PROTOCOLS[stim_type]
                ]
                for stim_type in stimuli_types
            }
        }
        # Extract the requested features for the requested protocols
        efeatures, protocol_definitions, _ = extract_efeatures(
            output_directory=temp_dir,
            files_metadata=files_metadata,
            targets=targets,
            absolute_amplitude=True,
            efel_settings=EFEL_SETTINGS,
        )
        output_features = {}
        logger.debug("Efeatures: %s", efeatures)
        # Format the extracted features into a readable dict for the model
        for protocol_name in protocol_definitions:
            efeatures_values = efeatures[protocol_name]
            protocol_def = protocol_definitions[protocol_name]
            output_features[protocol_name] = {
                f["efeature_name"]: {
                    "avg": f["val"][0],
                    "num_traces": f["n"],
                    "unit": get_unit(f["efeature_name"])
                    if get_unit(f["efeature_name"]) != "constant"
                    else None,
                }
                for f in efeatures_values["soma"]
            }

            # Add the stimulus current of the protocol to the output
            output_features[protocol_name]["stimulus_current"] = f"{protocol_def['step']['amp']} nA"
    return ElectrophysiologyMetricsOutput.from_efeatures(output_features)



from obi_one.scientific.morphology_locations import (
    ClusteredGroupedMorphologyLocations,
    ClusteredMorphologyLocations,
    ClusteredPathDistanceMorphologyLocations,
    MorphologyLocationsBlock,
    PathDistanceMorphologyLocations,
    RandomGroupedMorphologyLocations,
    RandomMorphologyLocations,
)

MorphologyLocationUnion = (
    MorphologyLocationsBlock
    | RandomGroupedMorphologyLocations
    | RandomMorphologyLocations
    | ClusteredGroupedMorphologyLocations
    | ClusteredMorphologyLocations
    | ClusteredPathDistanceMorphologyLocations
    | PathDistanceMorphologyLocations
)



import abc
from typing import Self

import numpy
from pydantic import Field, model_validator

from obi_one.core.block import Block

from .find_specified_afferent_synapses import (
    add_section_types,
    all_syns_on,
    apply_filters,
    merge_multiple_syns_per_connection,
    morphology_and_pathdistance_calculator,
    relevant_path_distances,
    select_by_path_distance,
    select_clusters_by_count,
    select_clusters_by_max_distance,
    select_minmax_distance,
    select_randomly,
)


class AfferentSynapsesBlock(Block, abc.ABC):
    """Base class representing the selection of afferent synapses according to specs."""

    random_seed: int | list[int] = Field(
        default=0, name="Random seed", description="Seed for the random selection of synapses"
    )
    section_types: None | tuple[int, ...] | list[tuple[int, ...]] = Field(
        default=None,
        name="Section types",
        description="Valid types of sections for synapses. 2: axon, 3: basal, 4: apical",
    )
    pre_synapse_class: None | str | list[str] = Field(
        default=None,
        name="Synapse class",
        description="Valid synapse classes. EXC: excitatory synapses; INH: inhibitory synapses",
    )
    consider_nan_pass: bool | list[bool] = Field(
        default=True,
        name="Consider nan to pass",
        description="If False, synapses with no 'synapse_class' pass, else not.",
    )
    pre_node_populations: None | tuple[str, ...] | list[tuple[str, ...]] = Field(
        default=None,
        name="Presynaptic populations",
        description="Names of presynaptic node populations to allow",
    )
    merge_multiple_syns_con: bool | list[bool] = Field(
        default=False,
        name="Merge multiple synapses per connection",
        description="""
        If True, multiple synapses from the same source neuron are merged by averaging.
        In this mode, it is not individual synapses that are selected, but presynaptic neurons.
        Where, if a neuron is selected, it is implied that all its synapses onto the target
        neuron are to be considered as selected.
        """,
    )

    def gather_synapse_info(self, circ, node_population, node_id):
        prop_filters = {}
        node_props = []
        if self.pre_synapse_class is not None:
            prop_filters["synapse_class"] = self.pre_synapse_class
            node_props.append("synapse_class")
        if self.pre_node_populations is not None:
            prop_filters["source_population"] = list(self.pre_node_populations)
        if self.section_types is not None:
            prop_filters["afferent_section_type"] = list(self.section_types)

        morph, PD = morphology_and_pathdistance_calculator(circ, node_population, node_id)
        syns = all_syns_on(circ, node_population, node_id, node_props)
        add_section_types(syns, morph)
        drop_nan = not self.consider_nan_pass
        syns = apply_filters(syns, prop_filters, drop_nan=drop_nan)
        soma_pds, pw_pds = relevant_path_distances(PD, syns)
        if self.merge_multiple_syns_con:
            syns, soma_pds, pw_pds = merge_multiple_syns_per_connection(syns, soma_pds, pw_pds)
        return syns, soma_pds, pw_pds

    @abc.abstractmethod
    def _select_syns(self, *args):
        """Returns a generated list of points for the morphology."""

    @abc.abstractmethod
    def _check_parameter_values(self):
        """Do specific checks on the validity of parameters."""

    @model_validator(mode="after")
    def check_parameter_values(self) -> Self:
        # Only check whenever list are resolved to individual objects
        self._check_parameter_values()
        return self

    def synapses_on(self, circ, node_population, node_id):
        self.enforce_no_lists()
        numpy.random.seed(self.random_seed)
        args = self.gather_synapse_info(circ, node_population, node_id)
        return self._select_syns(*args)


class RandomlySelectedNumberOfSynapses(AfferentSynapsesBlock):
    """Completely random synapses without constraint.
    Specified number picked without bias.
    """

    n: int | list[int] = Field(
        default=1,
        name="Number of synapses",
        description="Number of synapses to pick",
    )

    def _select_syns(self, syns, *args):
        return select_randomly(syns, n=self.n, raise_insufficient=False)

    def _check_parameter_values(self):
        if not isinstance(self.n, list):
            assert self.n > 0, "Number of synapses must be at least one!"


class RandomlySelectedFractionOfSynapses(AfferentSynapsesBlock):
    """Completely random synapses without constraint.
    Each picked with specified probability.
    """

    p: int | list[int] = Field(
        default=1.0,
        name="Fracton of synapses",
        description="Fracton of synapses to pick",
    )

    def _select_syns(self, syns, *args):
        return select_randomly(syns, p=self.p, raise_insufficient=False)

    def _check_parameter_values(self):
        if not isinstance(self.p, list):
            assert self.p > 0, "Fraction of synapses must be > 0!"
            assert self.p <= 1.0, "Number of synapses must be <= 1.0!"


class PathDistanceConstrainedNumberOfSynapses(RandomlySelectedNumberOfSynapses):
    """Pick from synapses between specified minimum and maximum path distances from the soma.
    Specified number of synapses within the path distance interval are picked without bias.
    """

    soma_pd_min: float | list[float] = Field(
        default=0.0,
        name="Minimum soma path distance",
        description="Minimum path distance in um to the soma for synapses",
    )
    soma_pd_max: float | list[float] = Field(
        default=1e12,
        name="Maximum soma path distance",
        description="Maximm path distance in um to the soma for synapses",
    )

    def _select_syns(self, syns, soma_pds, *args):
        return select_minmax_distance(
            syns,
            soma_pds,
            soma_pd_min=self.soma_pd_min,
            soma_pd_max=self.soma_pd_max,
            n=self.n,
            raise_insufficient=False,
        )


class PathDistanceConstrainedFractionOfSynapses(RandomlySelectedFractionOfSynapses):
    """Pick from synapses between specified minimum and maximum path distances from the soma.
    From the synapses within the path distance interval a specified fractio is picked.
    """

    soma_pd_min: float | list[float] = Field(
        default=0.0,
        name="Minimum soma path distance",
        description="Minimum path distance in um to the soma for synapses",
    )
    soma_pd_max: float | list[float] = Field(
        default=1e12,
        name="Maximum soma path distance",
        description="Maximm path distance in um to the soma for synapses",
    )

    def _select_syns(self, syns, soma_pds, *args):
        return select_minmax_distance(
            syns,
            soma_pds,
            soma_pd_min=self.soma_pd_min,
            soma_pd_max=self.soma_pd_max,
            n=self.p,
            raise_insufficient=False,
        )


class PathDistanceWeightedNumberOfSynapses(RandomlySelectedNumberOfSynapses):
    """Pick synapses with path distance-dependent bias. A specified number of synapses
    is picked, with synapses close to a specified path distance from the soma being
    more likely to be selected. The bias is expressed by a Gaussian mean and std.
    """

    soma_pd_mean: float | list[float] = Field(
        name="Mean soma path distance",
        description="Mean of a Gaussian for soma path distance in um for selecting synapses",
    )
    soma_pd_sd: float | list[float] = Field(
        name="SD for soma path distance",
        description="SD of a Gaussian for soma path distance in um for selecting synapses",
    )

    def _check_parameter_values(self):
        if not isinstance(self.soma_pd_sd, list):
            assert self.soma_pd_sd > 0, "SD of Gaussian must be > 0!"

    def _select_syns(self, syns, soma_pds, *args):
        return select_by_path_distance(
            syns,
            soma_pds,
            soma_pd_mean=self.soma_pd_mean,
            soma_pd_sd=self.soma_pd_sd,
            n=self.n,
            raise_insufficient=False,
        )


class PathDistanceWeightedFractionOfSynapses(RandomlySelectedFractionOfSynapses):
    """Pick synapses with path distance-dependent bias. A specified fraction of all synapses
    is picked, with synapses close to a specified path distance from the soma being
    more likely to be selected. The bias is expressed by a Gaussian mean and std.
    """

    soma_pd_mean: float | list[float] = Field(
        name="Mean soma path distance",
        description="Mean of a Gaussian for soma path distance in um for selecting synapses",
    )
    soma_pd_sd: float | list[float] = Field(
        name="SD for soma path distance",
        description="SD of a Gaussian for soma path distance in um for selecting synapses",
    )

    def _check_parameter_values(self):
        if not isinstance(self.soma_pd_sd, list):
            assert self.soma_pd_sd > 0, "SD of Gaussian must be > 0!"

    def _select_syns(self, syns, soma_pds, *args):
        return select_by_path_distance(
            syns,
            soma_pds,
            soma_pd_mean=self.soma_pd_mean,
            soma_pd_sd=self.soma_pd_sd,
            n=self.p,
            raise_insufficient=False,
        )


class ClusteredSynapsesByMaxDistance(AfferentSynapsesBlock):
    """Pick clusters of synapses. A cluster in this context comprises all synapses
    closer than a maximum path distance to a synapse that has been picked as a
    cluster center. The center is picked randomly without bias.
    """

    n_clusters: int | list[int] = Field(
        default=1, name="Number of clusters", description="Number of synapse clusters to find"
    )
    cluster_max_distance: float | list[float] = Field(
        name="Maximum distance of synapses from cluster center",
        description="Synapses within a cluster will be closer than this value from the cluster center (in um)",
    )

    def _check_parameter_values(self):
        if not isinstance(self.n_clusters, list):
            assert self.n_clusters > 0, "Must generate at least one cluster!"
        if not isinstance(self.cluster_max_distance, list):
            assert self.cluster_max_distance >= 0, "Cluster distance must be >= 0"

    def _select_syns(self, syns, soma_pds, pw_pds):
        return select_clusters_by_max_distance(
            syns,
            soma_pds,
            pw_pds,
            n_clusters=self.n_clusters,
            cluster_max_distance=self.cluster_max_distance,
            raise_insufficient=False,
        )


class ClusteredSynapsesByCount(AfferentSynapsesBlock):
    """Pick clusters of synapses. A cluster in this context comprises
    n_per_cluster synapses that are closest in path distance to a synapse that
    has been picked as a cluster center.
    The center is picked randomly without bias.
    """

    n_clusters: int | list[int] = Field(
        default=1, name="Number of clusters", description="Number of synapse clusters to find"
    )
    n_per_cluster: int | list[int] = Field(
        name="Number of synapses per cluster",
        description="This number of synapses per cluster will be selected by proximity to a center synapse.",
    )

    def _check_parameter_values(self):
        if not isinstance(self.n_clusters, list):
            assert self.n_clusters > 0, "Must generate at least one cluster!"
        if not isinstance(self.n_per_cluster, list):
            assert self.n_per_cluster > 0, "Must select at least one synapse per cluster!"

    def _select_syns(self, syns, soma_pds, pw_pds):
        return select_clusters_by_count(
            syns,
            soma_pds,
            pw_pds,
            n_clusters=self.n_clusters,
            n_per_cluster=self.n_per_cluster,
            raise_insufficient=False,
        )


class ClusteredPDSynapsesByMaxDistance(ClusteredSynapsesByMaxDistance):
    """Pick clusters of synapses. A cluster in this context comprises all synapses
    closer than a maximum path distance to a synapse that has been picked as a
    cluster center.
    The center is picked with a bias that depends on path distance to the soma.
    That is, synapse close to a specified path distance are more likely to be
    selected.
    """

    soma_pd_mean: float | list[float] = Field(
        name="Mean soma path distance",
        description="Mean of a Gaussian for soma path distance in um for selecting synapses",
    )
    soma_pd_sd: float | list[float] = Field(
        name="SD for soma path distance",
        description="SD of a Gaussian for soma path distance in um for selecting synapses",
    )

    def _check_parameter_values(self):
        if not isinstance(self.soma_pd_sd, list):
            assert self.soma_pd_sd > 0, "SD of Gaussian must be > 0!"

    def _select_syns(self, syns, soma_pds, pw_pds):
        return select_clusters_by_max_distance(
            syns,
            soma_pds,
            pw_pds,
            n_clusters=self.n_clusters,
            cluster_max_distance=self.cluster_max_distance,
            soma_pd_mean=self.soma_pd_mean,
            soma_pd_sd=self.soma_pd_sd,
            raise_insufficient=False,
        )


class ClusteredPDSynapsesByCount(ClusteredSynapsesByCount):
    """Pick clusters of synapses. A cluster in this context comprises
    n_per_cluster synapses that are closest in path distance to a synapse that
    has been picked as a cluster center.
    The center is picked with a bias that depends on path distance to the soma.
    That is, synapse close to a specified path distance are more likely to be
    selected.
    """

    soma_pd_mean: float | list[float] = Field(
        name="Mean soma path distance",
        description="Mean of a Gaussian for soma path distance in um for selecting synapses",
    )
    soma_pd_sd: float | list[float] = Field(
        name="SD for soma path distance",
        description="SD of a Gaussian for soma path distance in um for selecting synapses",
    )

    def _check_parameter_values(self):
        if not isinstance(self.soma_pd_sd, list):
            assert self.soma_pd_sd > 0, "SD of Gaussian must be > 0!"

    def _select_syns(self, syns, soma_pds, pw_pds):
        return select_clusters_by_count(
            syns,
            soma_pds,
            pw_pds,
            n_clusters=self.n_clusters,
            n_per_cluster=self.n_per_cluster,
            soma_pd_mean=self.soma_pd_mean,
            soma_pd_sd=self.soma_pd_sd,
            raise_insufficient=False,
        )



import abc
from typing import ClassVar

from entitysdk.models.entity import Entity
from pydantic import Field, PrivateAttr

import entitysdk

from obi_one.database.db_manager import db
from obi_one.core.base import OBIBaseModel


from enum import Enum
class LoadAssetMethod(Enum):
    MEMORY = "memory"
    FILE = "file"

class EntityFromID(OBIBaseModel, abc.ABC):
    entitysdk_class: ClassVar[type[Entity]] = None
    id_str: str = Field(description="ID of the entity in string format.")
    _entity: Entity | None = PrivateAttr(default=None)

    @classmethod
    def fetch(cls, entity_id: str, db_client: entitysdk.client.Client) -> Entity:
        return db_client.get_entity(
            entity_id=entity_id, entity_type=cls.entitysdk_class
        )

    @classmethod
    def find(cls, db_client: entitysdk.client.Client, limit: int = 10, **kwargs) -> list[Entity]:
        return db_client.search_entity(
            entity_type=cls.entitysdk_class, query=kwargs, limit=limit
        ).all()

    def entity(self, db_client: entitysdk.client.Client) -> Entity:
        if self._entity is None:
            self._entity = self.__class__.fetch(self.id_str, db_client=db_client)
        return self._entity

    @property
    def entitysdk_type(self) -> type[Entity]:
        return self.__class__.entitysdk_class



from obi_one.core.base import OBIBaseModel


class NamedPath(OBIBaseModel):
    """Helper class to assign a name to a file path."""

    name: str
    path: str

    def __repr__(self) -> str:
        """Return a string representation of the NamedPath object."""
        return self.name



import abc
from typing import Annotated

from pydantic import BaseModel, Field

from obi_one.core.activity import Activity


class Validation(Activity, abc.ABC):
    """Base class for validation objects.

    This class is used to define the structure of validation objects.
    It can be extended to create specific validation types.
    """


class SingleValidationOutput(BaseModel, abc.ABC):
    """Base class for single validation output objects.

    This class is used to define the base structure of validation output objects.
    """

    name: Annotated[
        str,
        Field(
            title="Validation Name",
            description="Name of the validation.",
            examples="Simulatable Neuron Depolarization Block Validation",
        )
    ]
    passed: Annotated[
        bool,
        Field(
            title="Validation Passed",
            description="Indicates whether the validation passed or not.",
            examples=True,
        )
    ]
    validation_details: Annotated[
        str,
        Field(
            title="Validation Details",
            description=(
                "Details about the validation, including any errors or warnings. "
                "It will be registgered as an asset on entitycore."
            ),
            examples="Validation passed: No depolarization block detected.."
        )
    ] = ""
    assets: Annotated[
        list[str],
        Field(
            title="Asset file paths",
            description=(
                "List of paths to the files to be registered as assets of the validation output."
            ),
            examples=["./MEM__1372346-C060114A5__cADpyr_L5_TPCA/depolarization_block_test.pdf"]
        )
    ] = []



from http import HTTPStatus
from typing import Annotated

import entitysdk.client
import entitysdk.exception
from fastapi import APIRouter, Depends, HTTPException, Query

from app.dependencies.entitysdk import get_client
from app.errors import ApiError, ApiErrorCode
from app.logger import L
from obi_one.scientific.ephys_extraction.ephys_extraction import (
    AmplitudeInput,
    CALCULATED_FEATURES,
    ElectrophysiologyMetricsOutput,
    STIMULI_TYPES,
    get_electrophysiology_metrics,
)
from obi_one.scientific.morphology_metrics.morphology_metrics import (
    MorphologyMetricsOutput,
    get_morphology_metrics,
)


def activate_declared_endpoints(router: APIRouter) -> APIRouter:
    @router.get(
        "/neuron-morphology-metrics/{reconstruction_morphology_id}",
        summary="Neuron morphology metrics",
        description="This calculates neuron morphology metrics for a given reconstruction \
                    morphology.",
    )
    def neuron_morphology_metrics_endpoint(
        db_client: Annotated[entitysdk.client.Client, Depends(get_client)],
        reconstruction_morphology_id: str,
    ) -> MorphologyMetricsOutput:
        L.info("get_morphology_metrics")

        try:
            metrics = get_morphology_metrics(
                reconstruction_morphology_id=reconstruction_morphology_id,
                db_client=db_client,
            )
        except entitysdk.exception.EntitySDKError:
            raise HTTPException(
                status_code=HTTPStatus.NOT_FOUND,
                detail={
                    "code": ApiErrorCode.NOT_FOUND,
                    "detail": (
                        f"Reconstruction morphology {reconstruction_morphology_id} not found."
                    ),
                },
            )

        if metrics:
            return metrics
        L.error(
            f"Reconstruction morphology {reconstruction_morphology_id} metrics computation issue"
        )
        raise ApiError(
            message="Asset not found",
            error_code=ApiErrorCode.NOT_FOUND,
            http_status_code=HTTPStatus.NOT_FOUND,
        )

    @router.get(
        "/electrophysiologyrecording-metrics/{trace_id}",
        summary="electrophysiology recording metrics",
        description="This calculates electrophysiology traces metrics for a particular recording",
    )
    def electrophysiologyrecording_metrics_endpoint(
        trace_id: str,
        db_client: Annotated[entitysdk.client.Client, Depends(get_client)],
        requested_metrics: CALCULATED_FEATURES | None = Query(default=None),
        amplitude: AmplitudeInput = Depends(),
        protocols: STIMULI_TYPES | None = Query(default=None),
    ) -> ElectrophysiologyMetricsOutput:
        try:
            ephys_metrics = get_electrophysiology_metrics(
                trace_id=trace_id,
                entity_client=db_client,
                calculated_feature=requested_metrics,
                amplitude=amplitude,
                stimuli_types=protocols,
            )
            return ephys_metrics
        except Exception as e:
            raise HTTPException(status_code=500, detail=f"Internal Server Error: {e}")

    return router




from obi_one.core.block import Block
from pydantic import Field

class Info(Block):
    campaign_name: str = Field(min_length=1, description="Name of the simulation campaign.")
    campaign_description: str = Field(min_length=1, description="Description of the simulation campaign.")


import logging
from pathlib import Path
from typing import Annotated, ClassVar

import morphio
import neurom
from pydantic import BaseModel, Field

from obi_one.core.validation import SingleValidationOutput, Validation

L = logging.getLogger(__name__)


class SingleReconstructionMorphologyValidationOutput(SingleValidationOutput):
    """Single output for a reconstruction morphology validation."""


class ReconstructionMorphologyValidationOutput(Validation):
    validation_a: Annotated[
        SingleReconstructionMorphologyValidationOutput,
        Field(
            title="validation_a",
            description="description of validation_a",
        ),
    ]
    validation_b: Annotated[
        SingleReconstructionMorphologyValidationOutput,
        Field(
            title="validation_b",
            description="description of validation_b",
        ),
    ]


class ReconstructionMorphologyValidation(Validation):
    """Validate the morphology of a reconstruction.

    This validation checks if the morphology of a reconstruction is valid.
    It is used to ensure that the morphology data meets certain criteria.
    """

    name: ClassVar[str] = "Validate Reconstruction Morphology"
    description: ClassVar[str] = "Validates the morphology of a reconstruction."
    morphology_file_path: Path | None = None

    _validation_output: ReconstructionMorphologyValidationOutput | None = None

    def run(self) -> None:
        """Run the validation logic."""
        L.info("Running Reconstruction Morphology Validation")

        if not self.morphology_file_path:
            raise ValueError("File path must be provided for validation.")

        neurom_morphology = neurom.load_morphology(self.morphology_file_path)
        morphio_morphology = morphio.Morphology(self.morphology_file_path)

        self._validation_output = ReconstructionMorphologyValidationOutput(
            validation_a=SingleReconstructionMorphologyValidationOutput(
                name="Morphology Validation A",
                passed=True,
                validation_details="Morphology is valid.",
            ),
            validation_b=SingleReconstructionMorphologyValidationOutput(
                name="Morphology Validation B",
                passed=False,
                validation_details="Axon section is missing.",
            ),
        )

        # Implement the validation logic here

    def save(self) -> None:
        """Save the result of the validation."""
        L.info("Saving Reconstruction Morphology Validation Output")

        if self._validation_output is None:
            raise ValueError("Validation output must be set before saving.")

        # Example: Save the validation output to a database or file



from typing import ClassVar

from obi_one.core.block import Block
from obi_one.core.form import Form
from obi_one.core.path import NamedPath
from obi_one.core.single import SingleCoordinateMixin


class MorphologyContainerizationsForm(Form):
    """ """

    single_coord_class_name: ClassVar[str] = "MorphologyContainerization"
    name: ClassVar[str] = "Morphology Containerization"
    description: ClassVar[str] = (
        "Creates a circuit with containerized morphologies instead of individual morphology files"
    )

    class Initialize(Block):
        circuit_path: NamedPath | list[NamedPath]
        hoc_template_old: str
        hoc_template_new: str

    initialize: Initialize


import datetime
import json
import logging
import os
import shutil
import traceback
from importlib.metadata import version
from typing import ClassVar

import h5py
import numpy as np
import tqdm
from bluepysnap import Circuit
from morph_tool import convert


class MorphologyContainerization(MorphologyContainerizationsForm, SingleCoordinateMixin):
    """Creates a circuit with containerized morphologies instead of individual morphology files,
    which involves the following steps:
    (1) Copy circuit to output location
    (2) Convert morphologies to .h5, if not yet existing (from .swc or .asc)
    (3) Merge individual .h5 morphologies into an .h5 container
    (4) Update the circuit config, pointing to the .h5 container
    (5) Update .hoc files so that they will work with .h5 containers
    (6) Delete all individual morphologies
    (7) Check containerized morphologies
    Important: The original circuit won't be modified! The circuit will be copied
               to the output location where all operations take place.
    """

    CONTAINER_FILENAME: ClassVar[str] = "merged-morphologies.h5"

    @staticmethod
    def _filter_ext(file_list, ext):
        """Filter file list based on file extension."""
        return list(filter(lambda f: os.path.splitext(f)[1].lower() == f".{ext}", file_list))

    @staticmethod
    def _check_morphologies(circuit_config):
        """Check modified circuit by loading some .h5 morphologies from each node population."""
        c = Circuit(circuit_config)
        for npop in c.nodes.population_names:
            nodes = c.nodes[npop]
            if nodes.type == "biophysical":
                for nid in nodes.ids()[[0, -1]]:  # First/last node ID
                    try:
                        morph = nodes.morph.get(
                            nid, transform=True, extension="h5"
                        )  # Will throw an error if not accessible
                    except:
                        return False  # Error
        return True  # All successful

    @staticmethod
    def _find_hoc_proc(proc_name, hoc_code):
        """Find a procedure with a given name in hoc code."""
        start_idx = hoc_code.find(f"proc {proc_name}")
        assert start_idx >= 0, f"ERROR: '{proc_name}' not found!"
        counter = 0
        has_first = False
        for _idx in range(start_idx, len(hoc_code)):
            if hoc_code[_idx] == "{":
                counter += 1
                has_first = True
            elif hoc_code[_idx] == "}":
                counter -= 1
            if has_first and counter == 0:
                end_idx = _idx
                break
        return start_idx, end_idx, hoc_code[start_idx : end_idx + 1]

    @staticmethod
    def _find_hoc_header(hoc_code):
        """Find the header section in hoc code."""
        start_idx = hoc_code.find("/*")  # First occurrence
        assert start_idx == 0, "ERROR: Header not found!"
        end_idx = hoc_code.find("*/")  # First occurrence
        assert end_idx > 0, "ERROR: Header not found!"
        return start_idx, end_idx, hoc_code[start_idx : end_idx + 2]

    def _update_hoc_files(self, hoc_folder):
        """Update hoc files in a folder from code of an old to code from a new template."""
        # TODO: CHECK IF .HOC FILE IS ALREADY NEW VERSION??
        # Extract code to be replaced from hoc templates
        with open(self.initialize.hoc_template_old) as f:
            tmpl_old = f.read()
        with open(self.initialize.hoc_template_new) as f:
            tmpl_new = f.read()

        proc_name = "load_morphology"
        _, _, hoc_code_old = self._find_hoc_proc(proc_name, tmpl_old)
        _, _, hoc_code_new = self._find_hoc_proc(proc_name, tmpl_new)

        # Replace code in hoc files
        for _file in tqdm.tqdm(os.listdir(hoc_folder), desc="Updating .hoc files"):
            if os.path.splitext(_file)[1].lower() != ".hoc":
                continue
            hoc_file = os.path.join(hoc_folder, _file)
            with open(hoc_file) as f:
                hoc = f.read()
            assert hoc.find(hoc_code_old) >= 0, "ERROR: Old HOC code to replace not found!"
            hoc_new = hoc.replace(hoc_code_old, hoc_code_new)
            _, _, header = self._find_hoc_header(hoc)
            module_name = self.__module__.split(".")[0]
            header_new = header.replace(
                "*/",
                f"Updated '{proc_name}' based on '{os.path.split(self.initialize.hoc_template_new)[1]}' by {module_name}({version(module_name)}) at {datetime.datetime.now()}\n*/",
            )
            hoc_new = hoc_new.replace(header, header_new)
            with open(hoc_file, "w") as f:
                f.write(hoc_new)

    def run(self) -> None:
        try:
            print(f"Running morphology containerization for '{self.initialize.circuit_path}'")

            # Set logging level to WARNING to prevent a lot of debug output from morph_tool.convert()
            logging.getLogger("morph_tool").setLevel(logging.WARNING)

            # Copy contents of original circuit folder to output_root
            input_path, input_config = os.path.split(self.initialize.circuit_path.path)
            output_path = self.coordinate_output_root
            circuit_config = os.path.join(output_path, input_config)
            assert not os.path.exists(circuit_config), "ERROR: Output circuit already exists!"
            print("Copying circuit to output folder...")
            shutil.copytree(input_path, output_path, dirs_exist_ok=True)
            print("...DONE")

            # Load circuit at new location
            c = Circuit(circuit_config)
            node_populations = c.nodes.population_names

            # Iterate over node populations to find all morphologies, convert them if needed, and merge them into a .h5 container
            hoc_folders_updated = []  # Keep track of updated folders (in case of different ones for different populations)
            morph_folders_to_delete = []  # Keep track of morphology folders (to be deleted afterwards)
            global_morph_entry = None  # Keep track wheter or not the circuit config has a global component entry for morphologies
            for npop in node_populations:
                nodes = c.nodes[npop]
                if nodes.type != "biophysical":
                    continue
                morph_names = np.unique(nodes.get(properties="morphology"))
                print(
                    f"> {len(morph_names)} unique morphologies in population '{npop}' ({nodes.size})"
                )

                # Check morphology folders
                morph_folders = {}
                for _morph_ext in ["h5", "asc", "swc"]:
                    try:
                        morph_folder = nodes.morph.get_morphology_dir(_morph_ext)
                        assert os.path.exists(morph_folder), (
                            f"ERROR: {_morph_ext} morphology folder does not exist!"
                        )
                        assert len(self._filter_ext(os.listdir(morph_folder), _morph_ext)) > 0, (
                            f"ERROR: {_morph_ext} morphology folder does not contain morphologies!"
                        )
                        if morph_folder not in morph_folders_to_delete:
                            morph_folders_to_delete.append(morph_folder)
                    except:
                        morph_folder = None
                    morph_folders[_morph_ext] = morph_folder

                # If .h5 morphologies not existing, run .asc/.swc to .h5 conversion
                h5_folder = morph_folders["h5"]
                if h5_folder is None:
                    for _morph_ext in ["asc", "swc"]:
                        inp_folder = morph_folders[_morph_ext]
                        if inp_folder is not None:
                            break
                    assert inp_folder is not None, "ERROR: No morphologies found to convert to .h5!"
                    h5_folder = os.path.join(os.path.split(inp_folder)[0], "_h5_morphologies_tmp_")
                    os.makedirs(h5_folder, exist_ok=True)

                    for _m in tqdm.tqdm(morph_names, desc=f"Converting .{_morph_ext} to .h5"):
                        src_file = os.path.join(inp_folder, _m + f".{_morph_ext}")
                        dest_file = os.path.join(h5_folder, _m + ".h5")
                        if not os.path.exists(dest_file):
                            convert(src_file, dest_file)

                # Merge into .h5 container
                if h5_folder not in morph_folders_to_delete:
                    morph_folders_to_delete.append(h5_folder)
                h5_container = os.path.join(os.path.split(h5_folder)[0], self.CONTAINER_FILENAME)
                with h5py.File(h5_container, "a") as f_container:
                    skip_counter = 0
                    for _m in tqdm.tqdm(morph_names, desc="Merging .h5 into container"):
                        with h5py.File(os.path.join(h5_folder, _m + ".h5")) as f_h5:
                            if _m in f_container:
                                skip_counter += 1
                            else:
                                f_h5.copy(f_h5, f_container, name=_m)
                print(
                    f"Merged {len(morph_names) - skip_counter} morphologies into container ({skip_counter} already existed)"
                )

                # Update the circuit config so that it points to the .h5 container file,
                # keeping the original global/local config file structure as similar as it was before
                # (but removing all other references to the original morphology folders)
                cname, cext = os.path.splitext(circuit_config)
                # shutil.copy(circuit_config, cname + "__BAK__" + cext)  # Save original config file

                with open(circuit_config) as f:
                    cfg_dict = json.load(f)

                if (
                    global_morph_entry is None
                ):  # Check if there is a global entry for morphologies (initially not set)
                    global_morph_entry = False
                    if "components" in cfg_dict:
                        if "morphologies_dir" in cfg_dict["components"]:
                            if len(cfg_dict["components"]["morphologies_dir"]) > 0:
                                base_path = os.path.split(
                                    cfg_dict["components"]["morphologies_dir"]
                                )[0]
                                cfg_dict["components"]["morphologies_dir"] = ""  # Remove .swc path
                                global_morph_entry = True
                        if "alternate_morphologies" in cfg_dict["components"]:
                            if (
                                "neurolucida-asc"
                                in cfg_dict["components"]["alternate_morphologies"]
                            ):
                                if (
                                    len(
                                        cfg_dict["components"]["alternate_morphologies"][
                                            "neurolucida-asc"
                                        ]
                                    )
                                    > 0
                                ):
                                    base_path = os.path.split(
                                        cfg_dict["components"]["alternate_morphologies"][
                                            "neurolucida-asc"
                                        ]
                                    )[0]
                                    cfg_dict["components"]["alternate_morphologies"][
                                        "neurolucida-asc"
                                    ] = ""  # Remove .asc path
                                    global_morph_entry = True
                            if "h5v1" in cfg_dict["components"]["alternate_morphologies"]:
                                if (
                                    len(cfg_dict["components"]["alternate_morphologies"]["h5v1"])
                                    > 0
                                ):
                                    base_path = os.path.split(
                                        cfg_dict["components"]["alternate_morphologies"]["h5v1"]
                                    )[0]
                                    cfg_dict["components"]["alternate_morphologies"]["h5v1"] = (
                                        ""  # Remove .h5 path
                                    )
                                    global_morph_entry = True
                        if global_morph_entry:
                            # Set .h5 container path globally
                            h5_file = os.path.join(base_path, self.CONTAINER_FILENAME)
                            cfg_dict["components"]["alternate_morphologies"] = {"h5v1": h5_file}

                if not global_morph_entry:  # Set individually per population
                    for _ndict in cfg_dict["networks"]["nodes"]:
                        if nodes.name in _ndict["populations"]:
                            _pop = _ndict["populations"][nodes.name]
                            base_path = None
                            if "morphologies_dir" in _pop:
                                if len(_pop["morphologies_dir"]) > 0:
                                    base_path = os.path.split(_pop["morphologies_dir"])[0]
                                    _pop["morphologies_dir"] = ""  # Remove .swc path
                            if "alternate_morphologies" in _pop:
                                if "neurolucida-asc" in _pop["alternate_morphologies"]:
                                    base_path = os.path.split(
                                        _pop["alternate_morphologies"]["neurolucida-asc"]
                                    )[0]
                                    _pop["alternate_morphologies"]["neurolucida-asc"] = (
                                        ""  # Remove .asc path
                                    )
                                if "h5v1" in _pop["alternate_morphologies"]:
                                    base_path = os.path.split(
                                        _pop["alternate_morphologies"]["h5v1"]
                                    )[0]
                                    _pop["alternate_morphologies"]["h5v1"] = ""  # Remove .h5 path
                            assert base_path is not None, (
                                f"ERROR: Morphology path for population '{nodes.name}' unknown!"
                            )
                            h5_file = os.path.join(base_path, self.CONTAINER_FILENAME)
                            _pop["alternate_morphologies"] = {"h5v1": h5_file}
                            break
                else:
                    pass  # Skip, should be already set

                with open(circuit_config, "w") as f:
                    json.dump(cfg_dict, f, indent=2)

                # Update hoc files (in place)
                hoc_folder = nodes.config["biophysical_neuron_models_dir"]
                if hoc_folder not in hoc_folders_updated:
                    self._update_hoc_files(hoc_folder)
                    hoc_folders_updated.append(hoc_folder)

            # Clean up morphology folders with individual morphologies
            print(f"Cleaning morphology folders: {morph_folders_to_delete}")
            for _folder in morph_folders_to_delete:
                shutil.rmtree(_folder)

            # Reload and check morphologies in modified circuit
            assert self._check_morphologies(circuit_config), (
                "ERROR: Morphology check not successful!"
            )
            print("Morphology containerization DONE")

        except Exception as e:
            traceback.print_exception(e)



from typing import Optional
import bluepysnap as snap
from conntility import ConnectivityMatrix
import numpy as np

from obi_one.core.base import OBIBaseModel


class Circuit(OBIBaseModel):
    """Class representing a circuit, i.e., pointing to a SONATA config and possible additional assets."""

    name: str
    path: str
    matrix_path: Optional[str] = None

    def __init__(self, name, path, **kwargs):
        super().__init__(name=name, path=path, **kwargs)
        _c = snap.Circuit(self.path)  # Basic check: Try to load the SONATA circuit w/o error

        if self.matrix_path is not None:
            _cmat = ConnectivityMatrix.from_h5(self.matrix_path)  # Basic check: Try to load the connectivity matrix w/o error
            np.testing.assert_array_equal(_cmat.vertices["node_ids"], _c.nodes[self._default_population_name(_c)].ids()) # TODO: This assumes the connectivity matrix is the local one, might nee to be extended in the future.

    def __str__(self):
        return self.name

    @property
    def sonata_circuit(self):
        """Provide access to SONATA circuit object."""
        return snap.Circuit(self.path)
    
    @property
    def connectivity_matrix(self):
        """Provide access to corresponding ConnectivityMatrix object."""
        if self.matrix_path is None:
            raise FileNotFoundError("Connectivity matrix has not been found")
        return ConnectivityMatrix.from_h5(self.matrix_path)

    @property
    def node_sets(self):
        """Returns list of available node sets."""
        return list(self.sonata_circuit.node_sets.content.keys())

    @staticmethod
    def get_node_population_names(c: snap.Circuit,incl_virtual=True):
        """Returns node population names."""
        popul_names = c.nodes.population_names
        if not incl_virtual:
            popul_names = [
                _pop for _pop in popul_names if c.nodes[_pop].type != "virtual"
            ]
        return popul_names
    
    @staticmethod
    def _default_population_name(c: snap.Circuit):
        """Returns the default node population name of a SONATA circuit c."""
        popul_names = Circuit.get_node_population_names(c, incl_virtual=False)
        assert len(popul_names) == 1, "Default node population unknown!"
        return popul_names[0]

    @property
    def default_population_name(self):
        """Returns the default node population name."""
        return self._default_population_name(self.sonata_circuit)

    def get_edge_population_names(self, incl_virtual=True):
        """Returns edge population names."""
        popul_names = self.sonata_circuit.edges.population_names
        if not incl_virtual:
            popul_names = [
                _pop
                for _pop in popul_names
                if self.sonata_circuit.edges[_pop].source.type != "virtual"
            ]
        return popul_names



"""Basic functions to compute network stats and for plotting
Last modified 03.2025
Author: Daniela Egas Santander
"""

import logging
import warnings

L = logging.getLogger(__name__)

import networkx as nx
import numpy as np
import pandas as pd

#Matplotlib imports
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.patches import FancyArrow
from matplotlib.patches import Ellipse
from matplotlib.gridspec import GridSpec
import matplotlib.patches as mpatches






try:
    from connalysis.network.classic import connection_probability_within, density
    from connalysis.network.topology import rc_submatrix, node_degree
except ImportError:
    warnings.warn("Connectome functionalities not available", UserWarning, stacklevel=1)


# Stats functions


def connection_probability_pathway(conn, grouping_prop):  # TODO: Add directly to connalysis?
    """Compute the connection probability of the matrix for a given grouping of the nodes"""

    def count_connections(mat, nrn):
        return mat.nnz

    def count_nodes(mat, nrn):
        return mat.shape

    # Setup analysis config per pathway
    analysis_specs = {
        "analyses": {
            "connection_counts": {
                "source": count_connections,
                "output": "scalar",
                "decorators": [
                    {
                        "name": "pathways_by_grouping_config",
                        "args": [{"columns": [grouping_prop], "method": "group_by_properties"}],
                    }
                ],
            },
            "node_counts": {
                "source": count_nodes,
                "output": "scalar",
                "decorators": [
                    {
                        "name": "pathways_by_grouping_config",
                        "args": [{"columns": [grouping_prop], "method": "group_by_properties"}],
                    }
                ],
            },
        }
    }
    out = conn.analyze(analysis_specs)

    # Compute connection probability
    df = out["node_counts"].unstack(f"idx-{grouping_prop}_post")
    diag = np.zeros(df.shape)
    np.fill_diagonal(diag, np.diag(df.map(lambda x: x[0]).to_numpy()))
    possible_connections = (df.map(lambda x: x[0] * x[1]) - diag).astype(int)
    connections = out["connection_counts"].unstack(f"idx-{grouping_prop}_post")
    connection_prob = connections / possible_connections
    return connection_prob


def connection_probability_within_pathway(conn, grouping_prop, max_dist=100):
    """Compute the connection probability within `max_dist` and for a given grouping of the nodes"""
    # Setup analysis config per pathway
    analysis_specs = {
        "analyses": {
            "probability_within": {
                "source": connection_probability_within,
                "args": [
                    ["x", "y", "z"],
                    max_dist,
                    "directed",
                ],  # [["x_um", "y_um", "z_um"], max_dist, "directed"],
                "output": "scalar",
                "decorators": [
                    {
                        "name": "pathways_by_grouping_config",
                        "args": [{"columns": [grouping_prop], "method": "group_by_properties"}],
                    }
                ],
            }
        }
    }
    out = conn.analyze(analysis_specs)
    return out["probability_within"].unstack(f"idx-{grouping_prop}_post")


def compute_global_connectivity(
    m,
    m_ER,
    v=None,
    type="full",
    max_dist=100,
    cols=["x", "y"],
):
    """Compute connection probabilities for the full network of with max_dist and similarly for the control"""
    if type == "full":  # Compute on the entire network
        return np.array(
            [density(m), density(m_ER), density(rc_submatrix(m)), density(rc_submatrix(m_ER))]
        )
    if type == "within":
        return np.array(
            [
                connection_probability_within(
                    m, v, max_dist=max_dist, cols=cols, type="directed", skip_symmetry_check=True
                ),
                connection_probability_within(
                    m_ER, v, max_dist=max_dist, cols=cols, type="directed", skip_symmetry_check=True
                ),
                connection_probability_within(
                    rc_submatrix(m),
                    v,
                    max_dist=max_dist,
                    cols=cols,
                    type="directed",
                    skip_symmetry_check=True,
                ),
                connection_probability_within(
                    rc_submatrix(m_ER),
                    v,
                    max_dist=max_dist,
                    cols=cols,
                    type="directed",
                    skip_symmetry_check=True,
                ),
            ]
        )


# Plotting functions


# Nodes
def make_pie_plot(ax, conn, grouping_prop, cmaps):
    category_counts = conn.vertices[grouping_prop].value_counts()
    category_counts = category_counts[category_counts > 0]

    # Group categories with percentages ≤ 2% into "Other"
    total = category_counts.sum()
    percentages = (category_counts / total) * 100
    small_categories = percentages[percentages <= 2].index
    if len(small_categories) > 1:
        other_count = category_counts[small_categories].sum()
        category_counts = category_counts.drop(small_categories)
        category_counts["Other"] = other_count

    # Define colors
    cmap = cmaps[grouping_prop]
    if grouping_prop == "synapse_class":
        # Fix red/blue for EXC/INH, if NA make it gray
        color_map = {"EXC": cmap(cmap.N), "INH": cmap(0), "NA":"gray"}
        colors = [color_map.get(key, cmap(i)) for i, key in enumerate(category_counts.index)]
    else:
        colors = [cmap(i) for i in range(len(category_counts))[::-1]]

    # Create the pie chart without percentages inside
    wedges, _ = ax.pie(category_counts, startangle=140, colors=colors, textprops={"fontsize": 8})

    # Add annotations outside the pie chart to avoid overlapping
    for i, wedge in enumerate(wedges):
        angle = (wedge.theta2 + wedge.theta1) / 2  # Midpoint angle of the wedge
        x = np.cos(np.radians(angle))  # X-coordinate for the label
        y = np.sin(np.radians(angle))  # Y-coordinate for the label
        extent = 1.4
        label_x = extent * x  # Position the label farther out
        label_y = extent * y
        ax.text(
            label_x,
            label_y,
            f"{category_counts.index[i]}: {percentages.iloc[i]:.1f}%",
            fontsize=8,
            ha="center",
            va="center",
        )

    # Adjust limits to ensure all labels are visible
    ax.set_xlim(-extent - 0.1, extent + 0.1)
    ax.set_ylim(-extent - 0.1, extent + 0.1)

    return ax


def plot_node_stats(conn, cmaps, full_width=17):
    fig = plt.figure(figsize=(full_width, full_width//3))
    gs = gridspec.GridSpec(2, 2, width_ratios=[1, 2.75])

    ax1 = fig.add_subplot(gs[0, 0])
    ax2 = fig.add_subplot(gs[1, 0])
    ax3 = fig.add_subplot(gs[:, 1])

    """Make plot of synapse class and mtype counts"""

    ax1.set_title("EI cell distribution")
    make_pie_plot(ax1, conn, "synapse_class", cmaps)
    ax2.set_title("Layer cell distribution")
    make_pie_plot(ax2, conn, "layer", cmaps)

    # mtype classes
    grouping_prop = "mtype"
    category_counts = conn.vertices[grouping_prop].value_counts()
    category_counts = category_counts[category_counts > 0]
    # Make bar chart
    cmap = cmaps[grouping_prop]
    category_counts.plot(kind="bar", color=cmap(cmap.N))
    ax3.set_xlabel("m-type")
    ax3.set_ylabel("Counts")
    ax3.set_title("m-type cell distribution")
    ax3.tick_params(axis="x", rotation=90)
    ax3.spines[["top", "right"]].set_visible(False)

    return fig


# Networks


def plot_degree(ax, deg, deg_ER, direction, type="full"):
    colors = ["teal", "lightgray"]
    for df, label, color in zip([deg, deg_ER], ["Connectome", "ER control"], colors, strict=False):
        df = df["IN"] + df["OUT"] if direction == "TOTAL" else df[direction]
        if type == "full":
            ax.plot(df.value_counts().sort_index(), label=label, color=color)
        elif type == "hist":
            ax.hist(df, alpha=0.5, label=label, color=color)
    return ax


from mpl_toolkits.axes_grid1.inset_locator import inset_axes


def plot_global_connection_probability(ax1, densities):
    # Connection probabilities
    colors = ["teal", "lightgrey", "teal", "lightgrey"]
    labels = ["Connectome", "ER control"]
    connectivity_label = ["Fulll", "Full", "Reciprocal", "Reciprocal"]
    hatches = ["", "", "//", "//"]  # Add stripes reciprocal connectivity

    # Plot full connectivity the primary y-axis
    bars1 = ax1.bar([0, 1], densities[:2], width=0.4, color=colors[:2])
    # ax1.legend(bars1, labels, frameon=False)

    # Create a secondary y-axis
    ax2 = ax1.twinx()
    # Plot reciprocal connectivity on the secondary y-axis
    bars2 = ax2.bar([2, 3], densities[2:], width=0.4, color=colors[2:])
    # Add hatches to reicprocal connectivity
    for bar, hatch in zip(bars2, hatches[2:], strict=False):
        bar.set_hatch(hatch)

    # Add labels to each bar
    ax1.set_xticks([0, 1, 2, 3], labels=connectivity_label)
    ax1.set_frame_on(False)
    ax2.set_frame_on(False)
    for bar, label in zip(bars1 + bars2, labels, strict=False):
        # height = bar.get_height()
        # ax = ax1 if bar in bars1 else ax2
        # ax.text(bar.get_x() + bar.get_width() / 2, height, label, ha='center', va='bottom')
        pass

    # Set labels and title
    # ax1.set_xlabel('Categories')
    ax1.set_ylabel("Connection probability")
    ax2.set_ylabel("Reciprocal connection probability", rotation=270, labelpad=20)
    ax1.ticklabel_format(style="scientific", axis="y", scilimits=(0, 0), useMathText=False)
    ax2.ticklabel_format(style="scientific", axis="y", scilimits=(0, 0), useMathText=False)
    # ax1.yaxis.set_major_formatter(FuncFormatter(lambda x, _: f'{x:.1e}'))
    #L.info(bars1)
    return ax1, bars1, labels


def plot_rc_connetion(ax, arrowsize=20, node_size=100):
    # Create a directed graph
    G = nx.DiGraph()
    G.add_node(1)
    G.add_node(2)
    G.add_edge(1, 2)
    G.add_edge(2, 1)

    # Draw the graph with curved edges
    pos = nx.circular_layout(G)
    nx.draw(
        G,
        pos,
        with_labels=False,
        node_color="black",
        edge_color="black",
        arrows=True,
        arrowsize=arrowsize,
        node_size=node_size,
        connectionstyle="arc3,rad=0.2",
        ax=ax,
    )
    return ax


def plot_in_out_deg(ax, direction, node_size=10, head_width=0.1, head_length=0.1, buffer=0.85):
    # Plot the central node
    ax.plot(0, 0, "ko", markersize=node_size)

    # Plot the arrows
    for i in range(5):
        angle = i * (360 / 5)
        x = 1.5 * np.cos(np.radians(angle))
        y = 1.5 * np.sin(np.radians(angle))
        if direction == "in":
            ax.arrow(
                x,
                y,
                -buffer * x,
                -buffer * y,
                head_width=head_width,
                head_length=head_length,
                fc="k",
                ec="k",
            )
        elif direction == "out":
            ax.arrow(
                0,
                0,
                buffer * x,
                buffer * y,
                head_width=head_width,
                head_length=head_length,
                fc="k",
                ec="k",
            )

    # Set the limits and aspect ratio
    ax.set_xlim(-2, 2)
    ax.set_ylim(-2, 2)
    ax.set_aspect("equal")
    ax.set_axis_off()
    return ax


def imshow_wrapper(ax, img, cutoff=15 * 15, perc=97.5, **kwargs):
    if np.prod(img.shape) > cutoff:
        kwargs.update(
            {"clim": [0.0, np.percentile(img.values.ravel()[~np.isnan(img.values.ravel())], perc)]}
        )
    plot = ax.imshow(img, **kwargs)
    return ax, plot


def plot_connection_probability_pathway(
    ax, connection_prob, cmap, cutoff=15 * 15, perc=97.5, **kwargs
):
    ax, plot = imshow_wrapper(ax, connection_prob, cutoff=cutoff, perc=perc, cmap=cmap, **kwargs)
    ax.set_yticks(range(len(connection_prob)), labels=connection_prob.index)
    ax.set_xticks(range(len(connection_prob)), labels=connection_prob.index)
    return ax, plot


def plot_connection_probability_stats(full_width, global_conn_probs):
    fig, axs = plt.subplots(
        1,
        5,
        figsize=(full_width, full_width // 3),
        gridspec_kw={"width_ratios": [1, 0.2, 1, 0.2, 0.6]},
    )
    axs[0].set_title("Connection probabilities overall", y=1.1, fontsize=14)
    axs[2].set_title("Connection probabilities within 100um", y=1.1, fontsize=14)

    # Global connection probabilities
    axs[0], bars, labels = plot_global_connection_probability(axs[0], global_conn_probs["full"])
    axs[2], bars, labels = plot_global_connection_probability(axs[2], global_conn_probs["widthin"])

    # Cartoons and labels
    ax = axs[4]
    inset_ax1 = inset_axes(ax, width="100%", height="20%", loc="upper left")
    inset_ax2 = inset_axes(
        ax,
        width="100%",
        height="40%",
        loc="center",
        bbox_to_anchor=(0.2, 0.2, 0.6, 0.8),
        bbox_transform=ax.transAxes,
    )
    inset_ax3 = inset_axes(ax, width="50%", height="50%", loc="lower left")
    inset_ax4 = inset_axes(ax, width="50%", height="50%", loc="lower right")

    ax.set_axis_off()  # Axis created just for white space

    inset_ax1.legend(
        bars, labels, frameon=False, ncol=2, loc="center"
    )  # , bbox_to_anchor=(0.25,1))
    inset_ax1.set_axis_off()  # Axis created just for white space

    plot_rc_connetion(inset_ax2, arrowsize=20, node_size=60)
    inset_ax2.set_title("Reciprocal \nconnection", fontsize=10, y=0.7)

    plot_in_out_deg(
        inset_ax3, direction="in", node_size=10, head_width=0.3, head_length=0.3, buffer=0.6
    )
    inset_ax3.set_title("In-degree", fontsize=10, y=0.8)

    plot_in_out_deg(
        inset_ax4, direction="out", node_size=10, head_width=0.3, head_length=0.3, buffer=0.6
    )
    inset_ax4.set_title("Out-degree", fontsize=10, y=0.8)

    for ax in [axs[1], axs[3]]:
        ax.set_axis_off()  # Axes created just for white space

    return fig


def plot_connection_probability_pathway_stats(full_width, conn_probs, deg, deg_ER):
    fig, axs = plt.subplots(3, 3, figsize=(full_width, full_width))

    for j, connection_type in enumerate(["full", "within"]):
        title = (
            "Connection probabilty \nper pathway overall"
            if connection_type == "full"
            else "Connection probabilty \nper pathway within 100um"
        )
        axs[0, j].text(
            0.5, 1.2, title, fontsize=14, ha="center", va="bottom", transform=axs[0, j].transAxes
        )

        # Connection probability
        for i, grouping_prop in enumerate(["synapse_class", "layer", "mtype"]):
            plotme = conn_probs[connection_type][grouping_prop]
            axs[i, j], plot = plot_connection_probability_pathway(axs[i, j], plotme, cmap="viridis")
            cbar = plt.colorbar(
                plot, ax=axs[i, j], orientation="vertical", shrink=0.85, label="Probability"
            )
            cbar.ax.ticklabel_format(style="scientific", axis="y", scilimits=(0, 0))
            axs[i, j].set_xlabel("Post-synaptic cell")
            axs[i, j].set_ylabel("Pre-synaptic cell")
        axs[2, j].set_title("Pathway: m-type")
        axs[2, j].tick_params(labelbottom=False, labelleft=False)
        axs[1, j].set_title("Pathway: layer")
        axs[0, j].set_title("Pathway: synapse class")

    # Plot degree distributions
    axs[0, 2].text(
        0.5,
        1.1,
        "Degree distributions",
        fontsize=14,
        ha="center",
        va="bottom",
        transform=axs[0, 2].transAxes,
    )
    for i, direction in enumerate(["IN", "OUT", "TOTAL"], start=0):
        axs[i, 2] = plot_degree(axs[i, 2], deg, deg_ER, direction, type="full")
        axs[i, 2].set_xlabel(f"{direction.capitalize()}-degree")
        axs[i, 2].spines[["top", "right"]].set_visible(False)
        axs[i, 2].set_frame_on(False)
        axs[i, 2].set_ylabel("Count")
        axs[i, 2].legend(frameon=False)

    fig.subplots_adjust(wspace=0.3)
    return fig

### Plotting function for small microcircuits 

def plot_smallMC_network_stats(conn, full_width,
                               color_indeg=plt.get_cmap("Set2")(0),
                               color_outdeg=plt.get_cmap("Set2")(2),
                               color_strength=plt.get_cmap("Set2")(1),
                               cmap_adj=plt.get_cmap("viridis")):
    
    fig, axs = plt.subplots(1,3, figsize=(full_width, full_width // 3))#, gridspec_kw={"width_ratios": [1, 2]})
    adj=conn.matrix
    adj_plot=adj.toarray().astype(float)
    adj_plot[adj_plot == 0] = np.nan
    # Connectivity matrix 
    min_val = int(np.nanmin(adj_plot[adj_plot > 0]))
    max_val = int(np.nanmax(adj_plot))
    plot=axs[0].imshow(adj_plot, cmap=cmap_adj, interpolation='nearest', aspect='auto', vmin=min_val, vmax=max_val)
    axs[0].xaxis.set_major_locator(plt.MaxNLocator(integer=True))
    axs[0].yaxis.set_major_locator(plt.MaxNLocator(integer=True))
    axs[0].set_xlabel("Target neuron ID")
    axs[0].set_ylabel("Source neuron ID")
    axs[0].set_title("Connectivity matrix")
    bbox = axs[0].get_position()
    cbar_height = 0.03  # Height of colorbar axis
    cbar_pad = 0.125     # Padding below the main axis (fraction of figure height)
    cbar_y = bbox.y0 - cbar_pad - cbar_height
    cax = fig.add_axes([bbox.x0, cbar_y, bbox.width, cbar_height])
    cbar = plt.colorbar(plot, cax=cax, orientation='horizontal', label="Synapse count")
    cbar.ax.xaxis.set_major_locator(plt.MaxNLocator(integer=True))

    # Synapse per connection
    unique_weights, counts = np.unique(adj.toarray(), return_counts=True)
    mask = (unique_weights != 0)
    unique_weights, counts = unique_weights[mask], counts[mask]
    axs[1].bar(unique_weights.astype(int), counts, color=color_strength)
    axs[1].xaxis.set_major_locator(plt.MaxNLocator(integer=True))
    if unique_weights.size == 1:
        axs[1].set_xticks(unique_weights)  # Otherwise it gives non-integer x-ticks
    axs[1].set_xlabel("Synapses per connection")
    axs[1].set_ylabel("Count")
    axs[1].set_title("Connection strength")

    #Plot degrees
    degree=node_degree(adj, direction=('IN', 'OUT'))
    bar_width = 0.4
    df = degree["IN"].value_counts().sort_index()
    axs[2].bar(df.index- bar_width/2 , df, width=bar_width, alpha=1, label="In degree", color=color_indeg)
    df = degree["OUT"].value_counts().sort_index()
    axs[2].bar(df.index+ bar_width/2 , df, width=bar_width, alpha=1, label="Out degree", color=color_outdeg)
    
    axs[2].xaxis.set_major_locator(plt.MaxNLocator(integer=True))
    unique_degrees=np.unique(degree)
    unique_degrees = unique_degrees[unique_degrees!=0]
    if unique_degrees.size == 1:
        axs[2].set_xticks(unique_degrees)  # Otherwise it gives non-integer x-ticks

    axs[2].set_xlabel("Degree")
    axs[2].set_ylabel("Count")
    axs[2].set_title("Node degrees")

    # Put legend below, otherwise sometimes it's over the bars
    bbox = axs[2].get_position()
    legend_height = 0.03  # Matching more or less the cbar options
    legend_pad = 0.125     # Matching more or less the cbar options
    legend_y = bbox.y0 - cbar_pad - cbar_height
    legend_ax = fig.add_axes([bbox.x0, legend_y, bbox.width, legend_height])
    handles, labels = axs[2].get_legend_handles_labels()
    legend_ax.legend(handles, labels, loc='center', frameon=False, ncol=2)
    legend_ax.axis('off')

    # Make axs[1] square 
    axs[0].set_box_aspect(1)
    axs[1].set_box_aspect(1)

    for ax in axs[1:]: 
        ax.spines[["top", "right"]].set_visible(False)

    return fig


def plot_growing_circles(fig, ax, radii, y1 = 0.5, color="black"): 
    # Get axis aspect ratio to make circles instead of ellipses
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width, height = bbox.width, bbox.height
    aspect = width / height    

    # Make even spacing between circles
    total_circle_width = sum([2 * r for r in radii])
    n = len(radii)
    gap = (1 - total_circle_width) / (n + 1)
    # Compute x positions 
    x1s = []
    x = gap + radii[0]
    for i in range(n):
        if i > 0:
            x += radii[i-1] + radii[i] + gap
        x1s.append(x)
    
    # Plot circles
    for i in range(n):
        radius = radii[i]
        x1 = x1s[i]
        radius_x = radius
        radius_y = radius * aspect
        ellipse = Ellipse((x1, y1), width=2*radius_x, height=2*radius_y, color=color, zorder=10, clip_on=False)
        ax.add_patch(ellipse)

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    ax.set_axis_off()

    return ax

def plot_growing_arrows(ax, widths, head_widths, y1=0.5, color="black", length=0.2, gap=0.05):
    n = len(widths)
    total_arrow_width = sum([length for _ in widths])
    total_gap = gap * (n + 1)
    total_width = total_arrow_width + total_gap
    start_x = (1 - total_width) / 2 + gap

    x1s = []
    x = start_x
    for i in range(n):
        x1s.append(x)
        x += length + gap

    for i in range(n):
        ax.add_patch(
            FancyArrow(
                x1s[i], y1, length, 0,
                width=widths[i]/100,
                head_width=head_widths[i],
                head_length=length/3,
                length_includes_head=True,
                color=color,
                zorder=10,
                clip_on=False
            )
        )

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    ax.set_axis_off()
    return ax


def plot_rc(ax, arrowsize=20, node_size=100):
    #Create graph 
    G = nx.DiGraph()
    G.add_node(0)
    G.add_node(1)
    G.add_edge(0, 1)
    G.add_edge(1, 0)

    # Draw the graph with curved edges
    pos = y=.5; x=0.1
    pos = {0: (x, y), 1: (x+0.5, y)}

    nx.draw(
        G,
        pos,
        with_labels=False,
        node_color="black",
        edge_color="black",
        arrows=True,
        arrowsize=arrowsize,
        node_size=node_size,
        connectionstyle="arc3,rad=0.2",
        ax=ax,
    )
    ax.set_box_aspect(1)
    return ax

def plot_small_network(ax, conn, 
                       node_color=None, edge_color=None, # To choose color of nodes and edges 
                       # To choose colors of nodes and edges by property, overrides node_color and edge_color
                       color_nodes_by_prop=False, color_map_nodes=None,color_property_nodes="synapse_class",
                       color_edges_by_prop=False, color_map_edges=None,color_property_edges="synapse_class",color_edges_by="pre",
                       edge_weight_scale=3, min_size= 300,  max_size =1500, 
                       title="Title!", title_fontsize=None,
                       projection="xy", coord_names=["x", "y"], 
                       axis_fontsize=None,
                       ): 


    ax.set_title(title)

    G=nx.from_numpy_array(conn.matrix.toarray(), create_using=nx.DiGraph)

    ## Choose position of neurons in 2D
    if projection =="xy":
        df=conn.vertices[coord_names]
        df["xy"] = df[coord_names].values.tolist()
        df = df.drop(columns=coord_names)
        pos=df.to_dict()['xy']
    elif projection =="circular": # Nodes in a circle
        pos = nx.circular_layout(G)               
    elif projection =="shell":
        pos = nx.shell_layout(G) # Nodes in concentric circles
    else: 
        raise ValueError(f"Projection type: {projection} not implemented. Choose from 'xy', 'circular', or 'shell'.")   

    ## Make edges proportional to weights
    weights = [G[u][v]['weight'] for u, v in G.edges()]
    widths = [w / max(weights) * edge_weight_scale for w in weights] # normalize for plotting

    ## Make nodes proportional to total degree
    total_degree=node_degree(conn.matrix.astype(bool).astype(int), direction=('IN', 'OUT')).sum(axis=1)
    min_deg, max_deg = min(total_degree), max(total_degree)
    if max_deg > min_deg:
        node_sizes = [min_size + (deg - min_deg) / (max_deg - min_deg) * (max_size - min_size) for deg in total_degree]
    else:
        node_sizes = [min_size for _ in total_degree]

    ## Color nodes and edges (possibly by property type) 
    if color_nodes_by_prop:
        node_colors = [color_map_nodes.get(key) for key in conn.vertices[color_property_nodes].to_numpy()]
    else: 
        node_colors=node_color
    if color_edges_by_prop:
        defining_colors = [color_map_edges.get(key) for key in conn.vertices[color_property_edges].to_numpy()]
        if color_edges_by == "pre":
            edge_colors = [defining_colors[u] for u, v in G.edges()]
        elif color_edges_by == "post":
            edge_colors = [defining_colors[v] for u, v in G.edges()]
        else: 
            raise ValueError(f"color_edges_by must be 'pre' or 'post', got {color_edges_by}.")
    else: 
        edge_colors=edge_color

    ## Plot network 
    nx.draw(G, pos, with_labels=True, 
            node_color=node_colors, edge_color=edge_colors,
            arrows=True, width=widths, node_size=node_sizes, 
            ax=ax)


    if projection=="xy":
        ## Add small axis to show it's a projection
        bbox = ax.get_position()

        # Coordinates for the mini-axis (in axes fraction of full axis)
        x0, y0 = 0.0, 0.0
        dx, dy = 0.1, 0.1

        # X axis
        ax.annotate('', xy=(x0 + dx, y0), xytext=(x0, y0),
                    arrowprops=dict(arrowstyle="->", lw=1, color='k'),
                    xycoords='axes fraction')
        ax.annotate(coord_names[0], xy=(x0 + dx, y0), xycoords='axes fraction', fontsize=axis_fontsize)

        # Y arrow
        ax.annotate('', xy=(x0, y0 + dy), xytext=(x0, y0),
                    arrowprops=dict(arrowstyle="->", lw=1, color='black'),
                    xycoords='axes fraction')

        ax.annotate(coord_names[1], xy=(x0, y0 + dy),xycoords='axes fraction', fontsize=axis_fontsize)
    return ax

def make_MC_fig_template(figsize, 
                         height_ratios=[1, 2, 1],  # relative row heights
                         width_ratios=[1, 1, 1], 
                         hspace_row1 = 0.05,
                         hspace_row2 = 0.15,
                         hspace_row3 = 0.02,  # hspaces between columns in each row (fraction)
                         cartoon_gaps = 0.1,  # gap between cartoons 
                         ax1_ratio = 0.5  # fraction of row width for ax1 

                         ):
    # Make template for figure of small MC network properties
    fig = plt.figure(figsize=figsize)

    # Define grid
    gs = GridSpec(3, 3,  # rows, columns
        height_ratios=height_ratios,  # row heights
        width_ratios=width_ratios,   # columns, will be used per row
        figure=fig
        )

    ### First row, connectivity and cartoons
    row1_bottom = gs[0, 0].get_position(fig).y0
    row1_top = gs[0, 0].get_position(fig).y1
    row1_height = row1_top - row1_bottom
    row1_left = gs[0, 0].get_position(fig).x0
    row1_right = gs[0, 2].get_position(fig).x1
    row1_width = row1_right - row1_left


    # ax1: left part, connectivity values
    col_width = (row1_width - hspace_row1) * ax1_ratio  

    ax1 = fig.add_axes([row1_left, row1_bottom, row1_left+col_width, row1_height])

    # ax2: right part, cartoons
    n_right = 3  # number of cartoons 
    right_start = row1_left + col_width + hspace_row1
    col_width = (row1_right-right_start-2*cartoon_gaps)/n_right
    ax2_1 = fig.add_axes([right_start, row1_bottom, col_width, row1_height])
    ax2_2 = fig.add_axes([right_start + col_width + cartoon_gaps, row1_bottom, col_width, row1_height])
    ax2_3 = fig.add_axes([right_start + 2 * (col_width + cartoon_gaps), row1_bottom, col_width, row1_height])
    ax2= (ax2_1, ax2_2, ax2_3)

    ### Second row
    row2_bottom = gs[1, 0].get_position(fig).y0
    row2_top = gs[1, 0].get_position(fig).y1
    row2_height = row2_top - row2_bottom
    row2_left = gs[1, 0].get_position(fig).x0
    row2_right = gs[1, 2].get_position(fig).x1
    row2_width = row2_right - row2_left
    col2_width = (row2_width - hspace_row2) / 2

    ax3 = fig.add_axes([row2_left, row2_bottom, col2_width, row2_height])
    ax4 = fig.add_axes([row2_left + col2_width + hspace_row2, row2_bottom, col2_width, row2_height])

    ### Third row
    row3_bottom = gs[2, 0].get_position(fig).y0
    row3_top = gs[2, 0].get_position(fig).y1
    row3_height = row3_top - row3_bottom
    row3_left = gs[2, 0].get_position(fig).x0
    row3_right = gs[2, 2].get_position(fig).x1
    row3_width = row3_right - row3_left
    col3_width = (row3_width - 2 * hspace_row3) / 3

    ax5 = fig.add_axes([row3_left, row3_bottom, col3_width, row3_height])
    ax6 = fig.add_axes([row3_left + col3_width + hspace_row3, row3_bottom, col3_width, row3_height])
    ax7 = fig.add_axes([row3_left + 2 * (col3_width + hspace_row3), row3_bottom, col3_width/2, row3_height])
    ax8 = fig.add_axes([row3_left + 2 * (col3_width + hspace_row3)+col3_width/2, row3_bottom, col3_width/2, row3_height])

    return fig, (ax1, ax2, ax3, ax4, ax5, ax6, ax7, ax8)


def plot_smallMC(conn, cmap, full_width ,textsize=14):
    ### Generate template for plot 
    fig, axs = make_MC_fig_template(figsize=(full_width, full_width),
                                    height_ratios=[1, 3, 0.3],  # relative row heights
                                    width_ratios=[1, 1, 1], 
                                    hspace_row1 = 0.15,
                                    hspace_row2 = 0.01,
                                    hspace_row3 = 0.01, # hspaces between columns in each row (fraction)
                                    cartoon_gaps = 0.01, # gap between cartoons 
                                    ax1_ratio = 0.3  # fraction of row width for ax1 
                                    )

    ax1, ax2, ax3, ax4, ax5, ax6, ax7, ax8=axs

    ### Plot connection probability 
    adj=conn.matrix.astype(bool).astype(int)
    x_pos=0.05
    ax1.text(x_pos, 0.7, f"Connection probability: {density(adj):.2e}", fontsize=textsize)
    ax1.text(x_pos, 0.3, f"Reciprocal connections (%): {density(rc_submatrix(adj))*100:.1f}%", fontsize=textsize)
    ax1.set_axis_off()

    ### Plot cartoons
    plot_rc(ax2[0], arrowsize=20, node_size=100)
    ax2[0].set_title("Reciprocal connection", fontsize=textsize, y=1)

    plot_in_out_deg(ax2[1], direction="in", node_size=10, head_width=0.3, head_length=0.3, buffer=0.6)
    ax2[1].set_title("In-degree", fontsize=textsize, y=1)

    plot_in_out_deg(ax2[2], direction="out", node_size=10, head_width=0.3, head_length=0.3, buffer=0.6)
    ax2[2].set_title("Out-degree", fontsize=textsize, y=1)

    ### Network plots
    # Color nodes by synapse class 
    color_map_nodes = {"INH": cmap(0), "EXC": cmap(cmap.N)}
    color_map_edges = {"INH": cmap(0), "EXC": cmap(cmap.N)}
    color_edges_by_prop=True
    color_property="synapse_class"

    # Plot x-y projection
    projection, coord_names, title ="xy", ["x", "y"], "Node positions: in x-y projection"
    #projection, coord_names, title ="circular", None, "Node positions: circular"
    ax3 = plot_small_network(ax3, conn, 
                            color_nodes_by_prop=True, color_map_nodes=color_map_nodes, color_property_nodes=color_property,
                            color_edges_by_prop=color_edges_by_prop, color_map_edges=color_map_edges, color_property_edges=color_property, color_edges_by="pre",
                            edge_weight_scale=4, min_size= 300,  max_size =1500, 
                            projection=projection, coord_names=coord_names, axis_fontsize=textsize,
                            title=title, title_fontsize=textsize)

    ### Plot circular projection
    projection, coord_names, title ="circular", None, "Node positions: circular"
    ax4 = plot_small_network(ax4, conn, 
                            color_nodes_by_prop=True, color_map_nodes=color_map_nodes, color_property_nodes=color_property,
                            color_edges_by_prop=color_edges_by_prop, color_map_edges=color_map_edges, color_property_edges=color_property, color_edges_by="pre",
                            edge_weight_scale=4, min_size= 300,  max_size =1500, 
                            projection=projection, coord_names=coord_names, axis_fontsize=textsize,
                            title=title, title_fontsize=textsize)

    # Node legends 
    largest_radius=0.1
    y_nodes=0.75
    # INH 
    color_map = {"INH": cmap(0), "EXC": cmap(cmap.N)} 
    plot_growing_circles(fig, ax7, radii = [largest_radius], y1 =y_nodes, color=color_map["EXC"][:-1]) 
    ax7.text(0.5, 0.1, "Excitatory neuron", va="center", ha="center", fontsize=12, color=color_map["EXC"][:-1])
    ax7.set_axis_off()
    # EXC
    plot_growing_circles(fig, ax8, radii = [largest_radius], y1 =y_nodes, color=color_map["INH"][:-1]) 
    ax8.text(0.5, 0.1, "Inhibitory neuron", va="center", ha="center", fontsize=12, color=color_map["INH"][:-1])
    ax8.set_axis_off()
    # Node size
    plot_growing_circles(fig, ax6, radii = [largest_radius/6,largest_radius/3, largest_radius/2], y1 = 0.75) 
    ax6.text(0.5, 0.1, "Node size represents in+out degree", va="center", ha="center", fontsize=14, color="black")

    # Edge width legend 
    plot_growing_arrows(ax5, widths=np.linspace(1,12,3), head_widths=[0.1, 0.2, 0.3], y1=0.75, color="black", length=0.2, gap=0.05)
    ax5.text(0.5, 0.1, "Edge thickness represents number of synapses", va="center", ha="center", fontsize=12, color="black")
    return fig

def plot_node_table(conn, figsize, 
                    colors_cmap, # name of discrete colormap from matplotlib
                    colors_file = None, # path to rgba colors file
                    h_scale = 2.5, 
                    v_scale = 2.5
                    ):
    """
    Plot a table of node properties with color coding.
    """

    # Get data frame of properties
    df = conn.vertices[["node_ids", "layer", "mtype"]]
    df = df.copy().rename(columns={"node_ids":"Neuron ID", "layer": "Layer", "mtype": "M-type"})
    df.insert(0, " ", [""] * len(df))  # Add empty column for circles

    # Get colors
    if colors_cmap != "custom": # From color map
        colors = plt.get_cmap(colors_cmap)
        N = conn._shape[0]
        if not (hasattr(colors, "colors") and colors.N >= N):
            raise ValueError(f"The rendering color map must contain at least as many colors as there are neurons.")
        colors = [colors(i) for i in range(colors.N)]
    else:  # Load colors from file 
        colors_df = pd.read_csv(colors_file, header=None)
        colors = [tuple(row) for row in colors_df.values]
        if not len(colors) >= len(df): 
            raise ValueError("The rendering color map must contain at least as many colors as there are neurons.")

    fig, ax = plt.subplots(figsize=figsize)
    ax.axis('off')
    ax.set_aspect('equal')  


    table = ax.table(
        cellText=df.values,
        colLabels=df.columns,
        loc='center',
        cellLoc='center'
    )
    table.auto_set_font_size(False)
    table.set_fontsize(12)
    table.scale(h_scale, v_scale)

    fig.canvas.draw()  # Draw table

    # Add color coding from the rendered image
    for i in range(len(df)):
        cell = table[i+1, 0]  # +1 for header row
        cell.get_text().set_text("")  # Remove text
        bbox = cell.get_window_extent(fig.canvas.get_renderer())
        inv = ax.transData.inverted()
        x0, y0 = inv.transform((bbox.x0, bbox.y0))
        x1, y1 = inv.transform((bbox.x1, bbox.y1))
        xc, yc = (x0 + x1) / 2, (y0 + y1) / 2
        radius = (y1 - y0) * 0.35
        circle = mpatches.Circle((xc, yc), radius, color=colors[i], zorder=10, clip_on=False)
        ax.add_patch(circle)
    
    return fig


from abc import ABC, abstractmethod
from obi_one.core.block import Block
from pydantic import Field, NonNegativeFloat
from typing import ClassVar

class SynapticManipulation(Block, ABC):

    @abstractmethod
    def _get_override_name(self) -> str:
        pass

    def config(self) -> dict:
        self.check_simulation_init()
        return self._generate_config()

    def _generate_config(self) -> dict:
        sonata_config = {
            "name": self._get_override_name(),
            "source": "All",
            "target": "All",
            "synapse_configure": self._get_synapse_configure(),
        }

        return sonata_config



class ScaleAcetylcholineUSESynapticManipulation(SynapticManipulation):
    """Applying a scaling factor to the U_SE parameter which determines the effect of achetylcholine (ACh) on synaptic release probability using the Tsodyks–Markram synaptic model.\
        This is applied for all synapses between biophysical neurons."""

    title: ClassVar[str] = "Demo: Scale U_SE to Modulate Acetylcholine Effect on Synaptic Release Probability"

    use_scaling: NonNegativeFloat | list[NonNegativeFloat] = Field(
        default=0.7050728631217412,
        title="Scale U_SE (ACh)",
        description="Scale the U_SE (ACh) parameter of the Tsodyks–Markram synaptic model.")

    def _get_override_name(self) -> str:
        return "ach_use"

    def _get_synapse_configure(self) -> str:
        return f"%s.Use *= {self.use_scaling}"


class SynapticMgManipulation(SynapticManipulation):
    """Manipulate the extracellular synaptic magnesium (Mg2+) concentration.\
        This is applied for all synapses between biophysical neurons."""

    title: ClassVar[str] = "Demo: Synaptic Mg2+ Concentration Manipulation"

    magnesium_value: NonNegativeFloat | list[NonNegativeFloat] = Field(
        default=2.4, 
        title="Extracellular Magnesium Concentration",
        description="Extracellular calcium concentration in millimoles (mM).", 
        units="mM"
    )
    
    def _get_override_name(self) -> str:
        return "Mg"

    def _get_synapse_configure(self) -> str:
        return f"%%s.mg = {self.magnesium_value}"



import time
import uuid
from unittest.mock import Mock

import httpx
import jwt
import pytest
from fastapi.security import HTTPAuthorizationCredentials

from app.config import settings
from app.dependencies import auth as test_module
from app.errors import ApiError, ApiErrorCode, AuthErrorReason
from app.schemas.auth import UserContext, UserContextWithProjectId
from app.schemas.base import OptionalProjectContext

from tests.utils import PROJECT_ID, UNRELATED_PROJECT_ID, UNRELATED_VIRTUAL_LAB_ID, VIRTUAL_LAB_ID

ZERO_UUID = "00000000-0000-0000-0000-000000000000"
TEST_USER_SUB = "4e234816-1391-4704-ab96-93f2c0875b42"
TEST_USER_EMAIL = "test@example.com"
TEST_USER_NAME = "John Doe"

PROJECT_CONTEXTS = [
    OptionalProjectContext(virtual_lab_id=None, project_id=None),
    OptionalProjectContext(virtual_lab_id=VIRTUAL_LAB_ID, project_id=PROJECT_ID),
]


@pytest.fixture(autouse=True)
def _clear_cache():
    yield
    test_module._check_user_info.cache_clear()


@pytest.fixture
def http_client():
    with httpx.Client() as client:
        yield client


@pytest.fixture
def request_mock(http_client):
    request = Mock()
    request.state.http_client = http_client
    return request


def _get_token(exp):
    decoded = {"sub": TEST_USER_SUB, "exp": exp, "email": TEST_USER_EMAIL}
    encoded = jwt.encode(decoded, "secret", algorithm="HS256")
    return HTTPAuthorizationCredentials(scheme="Bearer", credentials=encoded)


@pytest.mark.parametrize("project_context", PROJECT_CONTEXTS)
@pytest.mark.parametrize("is_token_jwt", [True, False])
@pytest.mark.parametrize("is_admin", [True, False])
@pytest.mark.usefixtures("freezer")
def test_user_verified_ok(httpx_mock, request_mock, is_admin, is_token_jwt, project_context):
    if is_token_jwt:
        token_expiration = int(time.time() + 3600)
        token = _get_token(exp=token_expiration)
    else:
        token_expiration = None
        token = HTTPAuthorizationCredentials(scheme="Bearer", credentials="random-ascii-token")

    httpx_mock.add_response(
        method="GET",
        url=test_module.KEYCLOAK_GROUPS_URL,
        match_headers={"Authorization": f"{token.scheme} {token.credentials}"},
        json={
            "sub": TEST_USER_SUB,
            "name": TEST_USER_NAME,
            "email": TEST_USER_EMAIL,
            "groups": [
                f"/service/{settings.APP_NAME}/admin" if is_admin else "/other",
                f"/proj/{VIRTUAL_LAB_ID}/{PROJECT_ID}/admin",
                f"/vlab/{VIRTUAL_LAB_ID}/admin",
            ],
        },
    )

    result = test_module.user_verified(
        project_context=project_context, token=token, request=request_mock
    )
    assert isinstance(result, UserContext)
    assert result == UserContext(
        subject=uuid.UUID(TEST_USER_SUB),
        email=TEST_USER_EMAIL,
        expiration=token_expiration,
        is_authorized=True,
        is_service_admin=is_admin,
        virtual_lab_id=project_context.virtual_lab_id,
        project_id=project_context.project_id,
        token=token,
    )


@pytest.mark.parametrize("project_context", PROJECT_CONTEXTS)
def test_user_verified_ok_when_auth_is_disabled(monkeypatch, request_mock, project_context):
    monkeypatch.setattr(settings, "APP_DISABLE_AUTH", True)
    result = test_module.user_verified(
        project_context=project_context, token=None, request=request_mock
    )
    assert isinstance(result, UserContext)
    assert result == UserContext(
        subject=uuid.UUID(ZERO_UUID),
        email=None,
        expiration=None,
        is_authorized=True,
        is_service_admin=True,
        virtual_lab_id=project_context.virtual_lab_id,
        project_id=project_context.project_id,
        token=None,
    )


def test_user_verified_without_token(request_mock):
    project_context = OptionalProjectContext(virtual_lab_id=None, project_id=None)
    with pytest.raises(ApiError) as excinfo:
        test_module.user_verified(project_context=project_context, token=None, request=request_mock)
    assert excinfo.value == ApiError(
        message=AuthErrorReason.AUTH_TOKEN_MISSING,
        error_code=ApiErrorCode.NOT_AUTHENTICATED,
        http_status_code=401,
    )


def test_user_verified_with_token_non_ascii(request_mock):
    project_context = OptionalProjectContext(virtual_lab_id=None, project_id=None)
    token = HTTPAuthorizationCredentials(scheme="Bearer", credentials="non-ascii-credential-é")
    with pytest.raises(ApiError) as excinfo:
        test_module.user_verified(
            project_context=project_context, token=token, request=request_mock
        )
    assert excinfo.value == ApiError(
        message=AuthErrorReason.AUTH_TOKEN_INVALID,
        error_code=ApiErrorCode.NOT_AUTHENTICATED,
        http_status_code=401,
    )


@pytest.mark.parametrize("project_context", PROJECT_CONTEXTS)
@pytest.mark.usefixtures("freezer")
def test_user_verified_with_expired_token(request_mock, project_context):
    token_expiration = int(time.time() - 1)
    token = _get_token(exp=token_expiration)

    with pytest.raises(ApiError) as excinfo:
        test_module.user_verified(
            project_context=project_context, token=token, request=request_mock
        )
    assert excinfo.value == ApiError(
        message=AuthErrorReason.AUTH_TOKEN_EXPIRED,
        error_code=ApiErrorCode.NOT_AUTHENTICATED,
        http_status_code=401,
    )


@pytest.mark.parametrize("project_context", PROJECT_CONTEXTS)
@pytest.mark.parametrize(
    ("keycloak_status", "expected_error"),
    [
        (
            401,
            ApiError(
                message=AuthErrorReason.NOT_AUTHENTICATED_USER,
                error_code=ApiErrorCode.NOT_AUTHENTICATED,
                http_status_code=401,
            ),
        ),
        (
            403,
            ApiError(
                message=AuthErrorReason.NOT_AUTHORIZED_USER,
                error_code=ApiErrorCode.NOT_AUTHORIZED,
                http_status_code=403,
            ),
        ),
        (
            500,
            ApiError(
                message="HTTP status error 500",
                error_code=ApiErrorCode.GENERIC_ERROR,
                http_status_code=500,
            ),
        ),
    ],
)
@pytest.mark.usefixtures("freezer")
def test_user_verified_keycloak_error(
    httpx_mock, request_mock, keycloak_status, expected_error, project_context
):
    token_expiration = int(time.time() + 3600)
    token = _get_token(exp=token_expiration)

    httpx_mock.add_response(
        status_code=keycloak_status,
        method="GET",
        url=test_module.KEYCLOAK_GROUPS_URL,
        match_headers={"Authorization": f"{token.scheme} {token.credentials}"},
    )

    with pytest.raises(ApiError) as excinfo:
        test_module.user_verified(
            project_context=project_context, token=token, request=request_mock
        )
    assert excinfo.value == expected_error


@pytest.mark.parametrize(
    # ("project_context", "expected_error"),
    "project_context",
    [
        OptionalProjectContext(virtual_lab_id=VIRTUAL_LAB_ID, project_id=PROJECT_ID),
        OptionalProjectContext(virtual_lab_id=None, project_id=PROJECT_ID),
        OptionalProjectContext(virtual_lab_id=VIRTUAL_LAB_ID, project_id=None),
    ],
)
@pytest.mark.usefixtures("freezer")
def test_user_verified_not_authorized_for_project(httpx_mock, request_mock, project_context):
    token_expiration = int(time.time() + 3600)
    token = _get_token(exp=token_expiration)

    httpx_mock.add_response(
        status_code=200,
        method="GET",
        url=test_module.KEYCLOAK_GROUPS_URL,
        match_headers={"Authorization": f"{token.scheme} {token.credentials}"},
        json={
            "sub": TEST_USER_SUB,
            "name": TEST_USER_NAME,
            "email": TEST_USER_EMAIL,
            "groups": [
                f"/service/{settings.APP_NAME}/admin",
                f"/proj/{UNRELATED_VIRTUAL_LAB_ID}/{UNRELATED_PROJECT_ID}/admin",
                f"/vlab/{UNRELATED_VIRTUAL_LAB_ID}/admin",
            ],
        },
    )

    with pytest.raises(ApiError) as excinfo:
        test_module.user_verified(
            project_context=project_context, token=token, request=request_mock
        )
    assert excinfo.value == ApiError(
        message=AuthErrorReason.NOT_AUTHORIZED_PROJECT,
        error_code=ApiErrorCode.NOT_AUTHORIZED,
        http_status_code=403,
    )


def test_user_with_project_id_ok():
    user_context = UserContext(
        subject=uuid.UUID(TEST_USER_SUB),
        email=TEST_USER_EMAIL,
        expiration=None,
        is_authorized=True,
        is_service_admin=True,
        virtual_lab_id=VIRTUAL_LAB_ID,
        project_id=PROJECT_ID,
        token=None,
    )

    result = test_module.user_with_project_id(user_context)
    assert isinstance(result, UserContextWithProjectId)
    assert result.virtual_lab_id == uuid.UUID(VIRTUAL_LAB_ID)
    assert result.project_id == uuid.UUID(PROJECT_ID)


@pytest.mark.parametrize(
    "project_context",
    [
        OptionalProjectContext(virtual_lab_id=None, project_id=None),
        OptionalProjectContext(virtual_lab_id=VIRTUAL_LAB_ID, project_id=None),
        OptionalProjectContext(virtual_lab_id=None, project_id=PROJECT_ID),
    ],
)
def test_user_with_project_id_raises(project_context):
    user_context = UserContext(
        subject=uuid.UUID(TEST_USER_SUB),
        email=TEST_USER_EMAIL,
        expiration=None,
        is_authorized=True,
        is_service_admin=True,
        virtual_lab_id=project_context.virtual_lab_id,
        project_id=project_context.project_id,
        token=None,
    )

    with pytest.raises(ApiError) as excinfo:
        test_module.user_with_project_id(user_context)
    assert excinfo.value == ApiError(
        message=AuthErrorReason.PROJECT_REQUIRED,
        error_code=ApiErrorCode.NOT_AUTHORIZED,
        http_status_code=403,
    )


def test_user_with_service_admin_role_ok():
    user_context = UserContext(
        subject=uuid.UUID(TEST_USER_SUB),
        email=TEST_USER_EMAIL,
        expiration=None,
        is_authorized=True,
        is_service_admin=True,
        virtual_lab_id=VIRTUAL_LAB_ID,
        project_id=PROJECT_ID,
        token=None,
    )
    result = test_module.user_with_service_admin_role(user_context)
    assert isinstance(result, UserContext)
    assert result.is_service_admin is True


def test_user_with_service_admin_role_raises():
    user_context = UserContext(
        subject=uuid.UUID(TEST_USER_SUB),
        email=TEST_USER_EMAIL,
        expiration=None,
        is_authorized=True,
        is_service_admin=False,
        virtual_lab_id=VIRTUAL_LAB_ID,
        project_id=PROJECT_ID,
        token=None,
    )
    with pytest.raises(ApiError) as excinfo:
        test_module.user_with_service_admin_role(user_context)
    assert excinfo.value == ApiError(
        message=AuthErrorReason.ADMIN_REQUIRED,
        error_code=ApiErrorCode.NOT_AUTHORIZED,
        http_status_code=403,
    )



from fastapi import FastAPI
from pydantic import BaseModel

from app.application import app
from obi_one.core import base


def test_app():
    assert isinstance(app, FastAPI)


def test_base_model():
    assert issubclass(base.OBIBaseModel, BaseModel)



import logging
import os
import traceback
import warnings
from typing import ClassVar, Optional
from pydantic import model_validator


import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path


L = logging.getLogger(__name__)

from obi_one.core.block import Block
from obi_one.core.form import Form
from obi_one.core.path import NamedPath
from obi_one.core.single import SingleCoordinateMixin
from obi_one.scientific.basic_connectivity_plots.helpers import (
    compute_global_connectivity,
    connection_probability_pathway,
    connection_probability_within_pathway,
    plot_connection_probability_pathway_stats,
    plot_connection_probability_stats,
    plot_node_stats,
    plot_smallMC_network_stats,
    plot_smallMC, 
    plot_node_table
)

import entitysdk.client

try:
    from connalysis.network.topology import node_degree
    from connalysis.randomization import ER_model
    from conntility import ConnectivityMatrix
except ImportError:
    warnings.warn("Connectome functionalities not available", UserWarning, stacklevel=1)


class BasicConnectivityPlots(Form):
    """Class to generate basic connectivity plots and stats from a ConnectivityMatrix object."""

    single_coord_class_name: ClassVar[str] = "BasicConnectivityPlot"
    name: ClassVar[str] = "Basic Connectivity Plots"
    description: ClassVar[str] = (
        "Generates basic connectivity plots and stats from a ConnectivityMatrix object."
    )

    class Initialize(Block):
        matrix_path: NamedPath | list[NamedPath]
        # TODO: implement node population option
        # node_population: None | str | list[None | str] = None
        plot_formats: tuple[str, ...] = ("png", "pdf", "svg")
        plot_types: tuple[str, ...] = ("nodes", # for any connectivity matrix
                                       "connectivity_global", "connectivity_pathway", # for medium and large connectomes
                                       "small_adj_and_stats", "network_in_2D", "property_table", # for small connectomes only
                                       )
        rendering_cmap: Optional[str] = None # Color map of the node identities
        rendering_color_file: Optional[str] = None # Color map file of the nod identities
        dpi: int = 300

        @model_validator(mode="after")
        def check_rendering_colors_for_property_table(self):
            if "property_table" in self.plot_types:
                if self.rendering_cmap == "custom":
                    if not Path(self.rendering_color_file).is_file():
                        raise ValueError(
                            "The rendering_color_file is not an existing file.")
                elif self.rendering_cmap is not None: 
                    cmap = plt.get_cmap(self.rendering_cmap)
                    if not hasattr(cmap, "colors"):
                        raise ValueError(f"You need to use a discrete color map")
                else: 
                    raise ValueError("When plotting `property_table` either a discrete colormap or a color map file must be passed.")
                
            return self

    initialize: Initialize


class BasicConnectivityPlot(BasicConnectivityPlots, SingleCoordinateMixin):
    """
    Generates and saves basic connectivity plots from a ConnectivityMatrix objects.

    Supported plot types:
      - "nodes": Node statistics (e.g., synapse class, layer, mtype).
      - "connectivity_pathway": Connection probabilities per pathway/grouping.  Not useful for small circuits.
      - "connectivity_global": Global connection probabilities across the network. Not useful for small circuits 
    - "small_adj_and_stats": Adjacency matrix and node statistics for small connectomes only (<= 20 nodes).
    - "network_in_2D": 2D visualization of the network for small connectomes only (<= 20 nodes).
    - "property_table": Table of node properties for small connectomes only (<= 20 nodes).

    Raises:
        Exception: If any error occurs during processing or plotting.
    """
    def run(self, db_client: entitysdk.client.Client = None) -> None:
        try:
            full_width = 16 #TODO: Maybe move this outside, but then fontsize would have to be changed accordingly
            # Set plot format, resolution and plot types
            plot_formats = self.initialize.plot_formats
            plot_types = self.initialize.plot_types
            dpi = self.initialize.dpi
            L.info(f"Plot Formats: {plot_formats}")
            L.info("Plot Types: {plot_types}")

            L.info(f"Info: Running idx {self.idx}, plots for {plot_types}")

            # Load matrix
            L.info(f"Info: Loading matrix '{self.initialize.matrix_path}'")
            conn = ConnectivityMatrix.from_h5(self.initialize.matrix_path.path)

            # Size metrics
            size = np.array([len(conn.vertices), conn.matrix.nnz, conn.matrix.sum()])
            L.info("Neuron, connection and synapse counts")
            L.info(size)
            output_file = os.path.join(self.coordinate_output_root, "size.npy")
            np.save(output_file, size)

            # Node metrics
            if "nodes" in plot_types:
                node_cmaps = {
                    "synapse_class": mcolors.LinearSegmentedColormap.from_list(
                        "RedBlue", ["C0", "C3"]
                    ),
                    "layer": plt.get_cmap("Dark2"),
                    "mtype": plt.get_cmap("GnBu"),
                }
                fig = plot_node_stats(conn, node_cmaps, full_width)
                for format in plot_formats:
                    output_file = os.path.join(self.coordinate_output_root, f"node_stats.{format}")
                    fig.savefig(output_file, dpi=dpi, bbox_inches="tight")

            # Degrees of matrix and control
            adj = conn.matrix.astype(bool)
            adj_ER = ER_model(adj)
            deg = node_degree(adj, direction=("IN", "OUT"))
            deg_ER = node_degree(adj_ER, direction=("IN", "OUT"))
            
            # Network metrics for large circuits
            # Connection probabilities per pathway
            if "connectivity_pathway" in plot_types:
                if size[0]<50: L.warning("Your network is likely too small for these plots to be informative.") 
                conn_probs = {"full": {}, "within": {}}
                for grouping_prop in ["synapse_class", "layer", "mtype"]:
                    conn_probs["full"][grouping_prop] = connection_probability_pathway(
                        conn, grouping_prop
                    )
                    conn_probs["within"][grouping_prop] = connection_probability_within_pathway(
                        conn, grouping_prop, max_dist=100
                    )
                # Plot network metrics
                fig_network_pathway = plot_connection_probability_pathway_stats(
                    full_width, conn_probs, deg, deg_ER
                )
                for format in plot_formats:
                    output_file = os.path.join(
                        self.coordinate_output_root, f"network_pathway_stats.{format}"
                    )
                    fig_network_pathway.savefig(output_file, dpi=dpi, bbox_inches="tight")

            # Global connection probabilities
            if "connectivity_global" in plot_types:
                if size[0]<50: L.warning("Your network is likely too small for these plots to be informative.") 
                # Global connection probabilities
                global_conn_probs = {"full": None, "within": None}
                global_conn_probs["full"] = compute_global_connectivity(adj, adj_ER, type="full")
                global_conn_probs["widthin"] = compute_global_connectivity(
                    adj, adj_ER, v=conn.vertices, type="within", max_dist=100, cols=["x", "y"]
                )

                # Plot network metrics
                fig_network_global = plot_connection_probability_stats(
                    full_width, global_conn_probs
                )
                for format in plot_formats:
                    output_file = os.path.join(
                        self.coordinate_output_root, f"network_global_stats.{format}"
                    )
                    fig_network_global.savefig(output_file, dpi=dpi, bbox_inches="tight")
            
            # Network metrics for small circuits
            # Plot the adjacency matrix, Nsyn and degrees
            if "small_adj_and_stats" in plot_types:
                if size[0]>20: 
                    L.warning("Your network is too large for these plots.") 
                else:
                    fig_adj_and_stats= plot_smallMC_network_stats(conn, full_width,
                                                                  color_indeg=plt.get_cmap("Set2")(0),
                                                                  color_outdeg=plt.get_cmap("Set2")(2),
                                                                  color_strength=plt.get_cmap("Set2")(1),
                                                                  cmap_adj=plt.get_cmap("viridis"))

                    for format in plot_formats:
                        output_file = os.path.join(
                            self.coordinate_output_root, f"small_adj_and_stats.{format}"
                        )
                        fig_adj_and_stats.savefig(output_file, dpi=dpi, bbox_inches="tight")

            # Plot network in 2D
            if "network_in_2D" in plot_types:
                if size[0]>20: 
                    L.warning("Your network is too large for these plots.") 
                else:
                    cmap = mcolors.LinearSegmentedColormap.from_list("RedBlue", ["C0", "C3"])
                    fig_network_in_2D= plot_smallMC(conn, cmap, full_width ,textsize=14)
                    
                    for format in plot_formats:
                        output_file = os.path.join(
                            self.coordinate_output_root, f"small_network_in_2D.{format}"
                        )
                        fig_network_in_2D.savefig(output_file, dpi=dpi, bbox_inches="tight")

            # Plot table of properties
            if "property_table" in plot_types:
                if size[0]>20: 
                    L.warning("Your network is too large for this table.") 
                else:
                    fig_property_table= plot_node_table(conn, figsize=(5,2), 
                    colors_cmap = self.initialize.rendering_cmap,
                    colors_file = self.initialize.rendering_color_file,
                    h_scale = 2.5, 
                    v_scale = 2.5
                    )
                    
                    for format in plot_formats:
                        output_file = os.path.join(
                            self.coordinate_output_root, f"property_table.{format}"
                        )
                        fig_property_table.savefig(output_file, dpi=dpi, bbox_inches="tight")

                    

            L.info(f"Done with {self.idx}")

        except Exception as e:
            traceback.print_exception(e)



import logging
from pathlib import Path
from typing import ClassVar

import morphio
import neurom.io
import neurom.view
import numpy as np
import pandas as pd
from fastapi import HTTPException
from matplotlib import pyplot as plt
from pydantic import Field

from obi_one.core.block import Block
from obi_one.core.form import Form
from obi_one.core.single import SingleCoordinateMixin
from obi_one.database.reconstruction_morphology_from_id import ReconstructionMorphologyFromID
from obi_one.scientific.unions.unions_morphology_locations import MorphologyLocationUnion

from .specified_morphology_locations import _PRE_IDX, _SEC_ID, _SEG_ID, _SEG_OFF

L = logging.getLogger(__name__)


class MorphologyLocationsForm(Form):
    """Form for generating locations on a morphology skeleton."""

    single_coord_class_name: ClassVar[str] = "MorphologyLocations"
    name: ClassVar[str] = "Point locations on neurite skeletons"
    description: ClassVar[str] = (
        "Generates optionally clustered locations on neurites of a morphology skeleton"
    )

    class Initialize(Block):
        morphology: (
            ReconstructionMorphologyFromID
            | list[ReconstructionMorphologyFromID]
            | Path
            | list[Path]
        ) = Field(title="Morphology", description="The morphology skeleton to place locations on")

    initialize: Initialize
    morph_locations: MorphologyLocationUnion = Field(
        title="Morphology locations",
        description="Parameterization of locations on the neurites of the morphology",
    )


class MorphologyLocations(MorphologyLocationsForm, SingleCoordinateMixin):
    """Generates locations on a morphology skeleton."""

    @staticmethod
    def generate_plot(m: morphio.Morphology, dataframe: pd.DataFrame) -> plt.figure:
        """Generate a plot of the morphology with locations on it."""

        def location_xyz(row: pd.Series) -> plt.figure:
            secid = int(row[_SEC_ID])
            segid = int(row[_SEG_ID])
            o = row[_SEG_OFF]
            seg = m.sections[secid - 1].points[segid : (segid + 2)]
            dseg = np.diff(seg, axis=0)[0]
            dseg = dseg / np.linalg.norm(dseg)
            return pd.Series(seg[0] + o * dseg, index=["x", "y", "z"])

        fig = plt.figure(figsize=(3, 6))
        ax = fig.gca()

        xyz = pd.concat([dataframe.apply(location_xyz, axis=1), dataframe[_PRE_IDX]], axis=1)
        neurom.view.plot_morph(neurom.io.utils.Morphology(m), ax=ax)
        xyz.groupby(_PRE_IDX).apply(lambda _x: ax.scatter(_x["x"], _x["y"], s=6))
        plt.axis("equal")
        return fig

    def run(self) -> None:
        try:
            if isinstance(self.initialize.morphology, Path):
                m = morphio.Morphology(self.initialize.morphology)
            else:
                m = self.initialize.morphology.morphio_morphology
            dataframe = self.morph_locations.points_on(m)

            fig = MorphologyLocations.generate_plot(m, dataframe)
            fig.savefig(self.coordinate_output_root / "locations_plot.pdf")
            dataframe.to_csv(self.coordinate_output_root / "morphology_locations.csv")

        except Exception as e:
            L.error(f"An error occurred: {e}")
            raise HTTPException(status_code=500, detail=f"Internal Server Error: {e}") from e



from pathlib import Path
from typing import ClassVar

import morphio
import neurom
import entitysdk
from entitysdk.models import Circuit
from entitysdk.models.entity import Entity
from entitysdk.staging.circuit import stage_circuit
from pydantic import PrivateAttr

from obi_one.database.db_manager import db
from obi_one.database.entity_from_id import EntityFromID, LoadAssetMethod

import io

class CircuitFromID(EntityFromID):
    entitysdk_class: ClassVar[type[Entity]] = Circuit
    _entity: Circuit | None = PrivateAttr(default=None)

    def download_circuit_directory(self, dest_dir=Path(), db_client: entitysdk.client.Client = None) -> None: 

        for asset in self.entity(db_client=db_client).assets:
            if asset.label == "sonata_circuit":

                circuit_dir = dest_dir / asset.path
                if circuit_dir.exists():
                    raise FileExistsError(f"Circuit directory '{circuit_dir}' already exists and is not empty.")

                circuit = db_client.get_entity(entity_id=self.id_str, entity_type=Circuit)
                stage_circuit(client=db_client, model=circuit, output_dir=circuit_dir, max_concurrent=4)

                break


from typing import Any

from pydantic import Field

from obi_one.core.base import OBIBaseModel


def nested_param_short(nested_param_list: list) -> str:
    """Convert a list of nested parameters to a short string representation."""
    nested_param_short = ""
    for i, s in enumerate(nested_param_list):
        nested_param_short += f"{s}"
        if i < len(nested_param_list) - 1:
            nested_param_short += "."
    return nested_param_short


class ScanParam(OBIBaseModel):
    location_list: list = Field(default_factory=list)
    _location_str: str = ""

    @property
    def location_str(self) -> str:
        """Return a string representation of the location list."""
        return nested_param_short(self.location_list)


class MultiValueScanParam(ScanParam):
    values: list[Any] = Field(default_factory=lambda: [None])


class SingleValueScanParam(ScanParam):
    value: Any



from typing import ClassVar

from pydantic import Field

from obi_one.core.block import Block
from obi_one.core.form import Form
from obi_one.core.path import NamedPath
from obi_one.core.single import SingleCoordinateMixin
from obi_one.database.reconstruction_morphology_from_id import ReconstructionMorphologyFromID


class SingleBlockGenerateTestForm(Form):
    """Test."""

    single_coord_class_name: ClassVar[str] = "SingleBlockGenerateTest"
    name: ClassVar[str] = "Single Block Generate Test"
    description: ClassVar[str] = "Test form for testing a single block form with entity SDK"

    class Initialize(Block):
        morphology_path: NamedPath | list[NamedPath]

    initialize: Initialize


class SingleBlockGenerateTest(SingleBlockGenerateTestForm, SingleCoordinateMixin):
    """Test."""

    @staticmethod
    def run() -> None:
        L.info("Running SingleBlockGenerateTest")


"""
Test Form for testing a single block form with entity SDK
"""


class SingleBlockEntityTestForm(Form):
    """Test Form for testing a single block form with entity SDK."""

    single_coord_class_name: ClassVar[str] = "SingleBlockGenerateTest"
    name: ClassVar[str] = "Single Block Entity Test"
    description: ClassVar[str] = "Test form for testing a single block form with entity SDK"

    class Initialize(Block):
        morphology: ReconstructionMorphologyFromID | list[ReconstructionMorphologyFromID]

    initialize: Initialize


class SingleBlockEntitySDKTest(SingleBlockEntityTestForm, SingleCoordinateMixin):
    """Test."""

    @staticmethod
    def run() -> None:
        L.info("Running SingleBlockEntitySDKTest")


class BlockForMultiBlockEntitySDKTest(Block):
    morphology_2: ReconstructionMorphologyFromID | list[ReconstructionMorphologyFromID]


class MultiBlockEntitySDKTestForm(Form):
    """Test."""

    single_coord_class_name: ClassVar[str] = "MultiBlockGenerateTest"
    name: ClassVar[str] = "Multi Block Entity Test"
    description: ClassVar[str] = "Test form for testing a single block form with entity SDK"

    test_blocks: dict[str, BlockForMultiBlockEntitySDKTest] = Field(description="Test blocks")

    class Initialize(Block):
        morphology: ReconstructionMorphologyFromID | list[ReconstructionMorphologyFromID]

    initialize: Initialize


class MultiBlockEntitySDKTest(MultiBlockEntitySDKTestForm, SingleCoordinateMixin):
    """Test."""

    @staticmethod
    def run() -> None:
        L.info("Running MultiBlockEntitySDKTest")



import io
import logging
from typing import Annotated, ClassVar

L = logging.getLogger(__name__)

import entitysdk
import neurom
from entitysdk.models.morphology import ReconstructionMorphology
from fastapi import HTTPException
from neurom import load_morphology
from pydantic import BaseModel, Field

from obi_one.core.block import Block
from obi_one.core.form import Form
from obi_one.core.single import SingleCoordinateMixin
from obi_one.database.reconstruction_morphology_from_id import ReconstructionMorphologyFromID


class MorphologyMetricsForm(Form):
    """ """

    single_coord_class_name: ClassVar[str] = "MorphologyMetrics"
    name: ClassVar[str] = "Morphology Metrics"
    description: ClassVar[str] = "Calculates morphology metrics for a given morphologies."

    class Initialize(Block):
        morphology: ReconstructionMorphologyFromID | list[ReconstructionMorphologyFromID] = Field(
            description="3. Morphology description"
        )

    initialize: Initialize

    def save(self, circuit_entities):
        """Add entitysdk calls to save the collection."""


class MorphologyMetricsOutput(BaseModel):
    aspect_ratio: Annotated[
        float,
        Field(
            title="aspect_ratio",
            description="Calculates the min/max ratio of the principal direction extents \
                along the plane.",
        ),
    ]
    circularity: Annotated[
        float,
        Field(
            title="circularity",
            description="Calculates the circularity of the morphology points along the plane.",
        ),
    ]
    length_fraction_above_soma: Annotated[
        float,
        Field(
            title="length_fraction_above_soma",
            description="Returns the length fraction of the segments that have their midpoints \
                            higher than the soma.",
        ),
    ]
    max_radial_distance: Annotated[
        float,
        Field(
            title="max_radial_distance",
            description="The maximum radial distance from the soma in micrometers.",
        ),
    ]
    number_of_neurites: Annotated[
        int, Field(title="number_of_neurites", description="Number of neurites in a morphology.")
    ]

    soma_radius: Annotated[
        float, Field(title="soma_radius [μm]", description="The radius of the soma in micrometers.")
    ]
    soma_surface_area: Annotated[
        float,
        Field(
            title="soma_surface_area [μm^2]",
            description="The surface area of the soma in square micrometers.",
        ),
    ]
    total_length: Annotated[
        float,
        Field(
            title="total_length [μm]",
            description="The total length of the morphology neurites in micrometers.",
        ),
    ]
    total_height: Annotated[
        float,
        Field(
            title="total_height [μm]",
            description="The total height (Y-range) of the morphology in micrometers.",
        ),
    ]
    total_height: Annotated[
        float,
        Field(
            title="total_width [μm]",
            description="The total width (X-range) of the morphology in micrometers.",
        ),
    ]
    total_depth: Annotated[
        float,
        Field(
            title="total_depth [μm]",
            description="The total depth (Z-range) of the morphology in micrometers.",
        ),
    ]
    total_area: Annotated[
        float,
        Field(
            title="total_area [μm^2]",
            description="The total surface area of all sections in square micrometers.",
        ),
    ]
    total_volume: Annotated[
        float,
        Field(
            title="total_volume [μm^3]",
            description="The total volume of all sections in cubic micrometers.",
        ),
    ]
    section_lengths: Annotated[
        list[float],
        Field(
            title="section_lengths [μm]",
            description="The distribution of lengths per section in micrometers.",
        ),
    ]
    segment_radii: Annotated[
        list[float],
        Field(
            title="segment_radii [μm]",
            description="The distribution of radii of the morphology in micrometers.",
        ),
    ]
    number_of_sections: Annotated[
        float,
        Field(
            title="number_of_sections",
            description="The number of sections in the morphology.",
        ),
    ]
    local_bifurcation_angles: Annotated[
        list[float],
        Field(
            title="local_bifurcation_angles [rad]",
            description="Angles between two sections computed at the bifurcation (local).",
        ),
    ]
    remote_bifurcation_angles: Annotated[
        list[float],
        Field(
            title="remote_bifurcation_angles [rad]",
            description="Angles between two sections computed at the end of the sections (remote).",
        ),
    ]
    section_path_distances: Annotated[
        list[float],
        Field(
            title="section_path_distances [μm]",
            description="Path distances from the soma to section endpoints in micrometers.",
        ),
    ]
    section_radial_distances: Annotated[
        list[float],
        Field(
            title="section_radial_distances [μm]",
            description="Radial distance from the soma to section endpoints in micrometers.",
        ),
    ]
    section_branch_orders: Annotated[
        list[float],
        Field(
            title="section_branch_orders",
            description="The distribution of branch orders of sections, computed from soma.",
        ),
    ]
    section_strahler_orders: Annotated[
        list[float],
        Field(
            title="section_strahler_orders",
            description="The distribution of strahler branch orders of sections, computed from terminals.",
        ),
    ]
	
	



    @classmethod
    def from_morphology(cls, neurom_morphology):
        return cls(
            aspect_ratio=neurom.get("aspect_ratio", neurom_morphology),
            circularity=neurom.get("circularity", neurom_morphology),
            length_fraction_above_soma=neurom.get("length_fraction_above_soma", neurom_morphology),
            max_radial_distance=neurom.get("max_radial_distance", neurom_morphology),
            number_of_neurites=neurom.get("number_of_neurites", neurom_morphology),
            soma_radius=neurom.get("soma_radius", neurom_morphology),
            soma_surface_area=neurom.get("soma_surface_area", neurom_morphology),
            total_length=neurom.get("total_length", neurom_morphology),
            total_height=neurom.get("total_height", neurom_morphology),
            total_width=neurom.get("total_width", neurom_morphology),
            total_depth=neurom.get("total_depth", neurom_morphology),
            total_area=neurom.get("total_area", neurom_morphology),
            total_volume=neurom.get("total_volume", neurom_morphology),
            section_lengths=neurom.get("section_lengths", neurom_morphology),
            segment_radii=neurom.get("segment_radii", neurom_morphology),
            number_of_sections=neurom.get("number_of_sections", neurom_morphology),
            local_bifurcation_angles=neurom.get("local_bifurcation_angles", neurom_morphology),
            remote_bifurcation_angles=neurom.get("remote_bifurcation_angles", neurom_morphology),
            section_path_distances=neurom.get("section_path_distances", neurom_morphology),
            section_radial_distances=neurom.get("section_radial_distances", neurom_morphology),
            section_branch_orders=neurom.get("section_branch_orders", neurom_morphology),
            section_strahler_orders=neurom.get("section_strahler_orders", neurom_morphology),
        )



class MorphologyMetrics(MorphologyMetricsForm, SingleCoordinateMixin):
    def run(self, db_client: entitysdk.client.Client = None):
        try:
            print("Running Morphology Metrics...")
            morphology_metrics = MorphologyMetricsOutput.from_morphology(
                    self.initialize.morphology.neurom_morphology(db_client=db_client)
                )
            L.info(morphology_metrics)

            return morphology_metrics

        except Exception as e:  # noqa: BLE001
            raise HTTPException(status_code=500, detail=f"Internal Server Error: {e}")


def get_morphology_metrics(
                reconstruction_morphology_id: str, 
                db_client: entitysdk.client.Client
            ) -> MorphologyMetricsOutput:

    morphology = db_client.get_entity(
        entity_id=reconstruction_morphology_id, 
        entity_type=ReconstructionMorphology
    )

    # Iterate through the assets of the morphology to find the one with content
    # type "application/asc"
    for asset in morphology.assets:
        if asset.content_type == "application/swc":
            # Download the content into memory
            content = db_client.download_content(
                entity_id=morphology.id,
                entity_type=ReconstructionMorphology,
                asset_id=asset.id,
            ).decode(encoding="utf-8")

            # Use StringIO to create a file-like object in memory from the string content
            neurom_morphology = load_morphology(io.StringIO(content), reader="swc")

            # Calculate the metrics using neurom
            morphology_metrics = MorphologyMetricsOutput.from_morphology(neurom_morphology)

            return morphology_metrics
    return None



import json
from pathlib import Path
from typing import Any

from entitysdk.models.entity import Entity

import obi_one as obi


def entity_encoder(obj: Any) -> dict[str, str]:
    """Encode an Entity into a JSON-serializable dictionary."""
    cls_name = obj.__class__.__name__
    if issubclass(obj.__class__, Entity) and "FromID" not in cls_name:
        return {"type": f"{cls_name}FromID", "id_str": str(obj.id)}
    if "FromID" in cls_name:
        return {"type": cls_name, "id_str": str(obj.id)}
    msg = f"Object of type {cls_name} is not JSON serializable"
    raise TypeError(msg)


def deserialize_obi_object_from_json_data(json_dict: dict) -> obi.OBIBaseModel:
    obi_object = getattr(obi, json_dict["type"]).model_validate(json_dict)
    return obi_object


def deserialize_obi_object_from_json_file(json_path: Path) -> obi.OBIBaseModel:
    with Path.open(json_path) as file:
        json_dict = json.load(file)
    return deserialize_obi_object_from_json_data(json_dict)



from obi_one.scientific.circuit.neuron_sets import (
    CombinedNeuronSet,
    IDNeuronSet,
    PredefinedNeuronSet,
    PropertyNeuronSet,
    VolumetricCountNeuronSet,
    VolumetricRadiusNeuronSet,
    SimplexNeuronSet,
    SimplexMembershipBasedNeuronSet,
    nbS1VPMInputs,
    nbS1POmInputs,
    rCA1CA3Inputs,
    AllNeurons,
    ExcitatoryNeurons,
    InhibitoryNeurons,
)

NeuronSetUnion = (
    PredefinedNeuronSet
    | CombinedNeuronSet
    | IDNeuronSet
    | PropertyNeuronSet
    | VolumetricCountNeuronSet
    | VolumetricRadiusNeuronSet
    | SimplexNeuronSet
    | SimplexMembershipBasedNeuronSet
    | nbS1VPMInputs
    | nbS1POmInputs
    | rCA1CA3Inputs
    | AllNeurons
    | ExcitatoryNeurons
    | InhibitoryNeurons
)

SimulationNeuronSetUnion = (
    AllNeurons
    | ExcitatoryNeurons
    | InhibitoryNeurons
    | IDNeuronSet
    | nbS1VPMInputs
    | nbS1POmInputs
    | rCA1CA3Inputs
)

from obi_one.core.block_reference import BlockReference
from typing import ClassVar, Any
class NeuronSetReference(BlockReference):
    """A reference to a NeuronSet block."""
    
    allowed_block_types: ClassVar[Any] = NeuronSetUnion


# class SimulationNeuronSetReference(BlockReference):
#     """A reference to a NeuronSet block for simulation."""
    
#     allowed_block_types: ClassVar[Any] = SimulationNeuronSetUnion